from dbxio.blobs.block_upload import upload_file
from dbxio.blobs.download import download_blob_tree
from dbxio.blobs.parquet import create_pa_table, create_tmp_parquet, pa_table2parquet

__all__ = [
    'create_pa_table',
    'create_tmp_parquet',
    'pa_table2parquet',
    'upload_file',
    'download_blob_tree',
]



import concurrent.futures
import time
import uuid
from abc import ABCMeta, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Optional, Union

import pandas as pd
import pyarrow as pa
from databricks.sdk.service.sql import (
    ExecuteStatementResponse,
    StatementExecutionAPI,
    StatementState,
    StatementStatus,
)
from databricks.sql.client import Connection, Cursor

from dbxio.blobs.parquet import arrow_stream2parquet, pa_table2parquet
from dbxio.core.exceptions import ReadDataError
from dbxio.utils.http import get_session
from dbxio.utils.logging import get_logger

STATEMENT_API_TIMEOUT = 3 * 60 * 60  # 3 hours
SLEEP_TIME = 10

ODBC_BATCH_SIZE_TO_FETCH = 100000

logger = get_logger()


class _FutureBaseResult(metaclass=ABCMeta):
    @abstractmethod
    def __enter__(self):
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    @abstractmethod
    def __next__(self) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def wait(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def full_results_path(self, results_path: Path) -> Union[Path, None]:
        raise NotImplementedError()

    @abstractmethod
    def download_and_save(self, results_path: str, max_concurrency: int = 1) -> Path:
        raise NotImplementedError()

    def df(self) -> pd.DataFrame:
        return pd.DataFrame(self)


class _FutureODBCResult(_FutureBaseResult):
    def __init__(self, future: concurrent.futures.Future):
        self._future = future

        self.statement_id = str(uuid.uuid4())
        logger.info('Statement ID: %s', self.statement_id)

        self._cached_batch: Optional[list[dict]] = None
        self._cached_batch_idx = 0

        self._batch_index = 0

    def full_results_path(self, results_path: Path) -> Path:
        return results_path / self.statement_id

    def download_and_save(self, results_path: str, max_concurrency: int = 1) -> Path:
        path = Path(results_path)
        assert path.exists(), f'{results_path} does not exist'

        self.wait()

        data_save_path = self.full_results_path(path)
        data_save_path.mkdir(parents=True, exist_ok=True)
        while True:
            try:
                table = self._fetch_batch()
                with open(data_save_path / f'{self._batch_index}.parquet', 'wb') as f:
                    f.write(pa_table2parquet(table))

                self._batch_index += 1
            except StopIteration:
                break

        return data_save_path

    @cached_property
    def _result(self) -> tuple[Connection, Cursor]:
        return list(concurrent.futures.as_completed([self._future]))[0].result()

    @cached_property
    def conn(self) -> Connection:
        return self._result[0]

    @cached_property
    def cursor(self) -> Cursor:
        return self._result[1]

    def __iter__(self):
        concurrent.futures.wait([self._future])
        return self

    def _fetch_batch(self) -> pa.Table:
        table = self.cursor.fetchmany_arrow(ODBC_BATCH_SIZE_TO_FETCH)
        if table.num_rows > 0:
            return table

        raise StopIteration

    def __next__(self) -> dict:
        if self._cached_batch and self._cached_batch_idx < len(self._cached_batch):
            record = self._cached_batch[self._cached_batch_idx]
            self._cached_batch_idx += 1
            return record
        else:
            self._cached_batch = self._fetch_batch().to_pandas().to_dict('records')
            self._cached_batch_idx = 0
            return self.__next__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cursor.close()
        self.conn.close()

    def wait(self) -> None:
        self._future.result()


class _FutureStatementApiResult(_FutureBaseResult):
    def __init__(self, response: ExecuteStatementResponse, statement_api: StatementExecutionAPI):
        self.statement_api = statement_api
        self._requests_session = get_session()

        assert response.statement_id, 'Statement ID is not found in the response'
        self.statement_id = response.statement_id

        self._cached_batch: list[dict] = []
        self._cached_batch_idx = 0

        self._total_chunk_count: int = (response.manifest.total_chunk_count if response.manifest else 0) or 0
        self._chunk_ptr = 0

    def full_results_path(self, results_path: Path) -> Path:
        return results_path / self.statement_id

    def download_and_save(self, results_path: str, max_concurrency: int = 1) -> Path:
        path = Path(results_path)
        assert path.exists(), f'{results_path} does not exist'
        self.wait()

        data_save_path = self.full_results_path(path)
        data_save_path.mkdir(parents=True, exist_ok=True)

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_concurrency) as executor:
            executor.map(
                lambda args: self._download_chunk(*args),
                [(chunk_idx, data_save_path) for chunk_idx in range(self._total_chunk_count)],
            )

        self._chunk_ptr = self._total_chunk_count  # to avoid downloading chunks again

        return data_save_path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def raise_for_status(self, status: StatementStatus = None):
        if status is None:
            status = self.statement_api.get_statement(self.statement_id).status
            assert status, 'Statement status is not found'
        state = status.state
        assert state, 'Statement state is not found'

        if state == StatementState.CANCELED:
            msg = f'Statement {self.statement_id} was canceled'
            logger.warning(msg)
            raise ReadDataError(msg)

        if state == StatementState.CLOSED:
            msg = f'Statement {self.statement_id} was closed. result no longer available for fetch'
            logger.warning(msg)
            raise ReadDataError(msg)

        if state == StatementState.FAILED:
            msg = f'Statement {self.statement_id} failed. ' f'Error ({status.error.error_code}): {status.error.message}'  # type: ignore
            logger.error(msg)
            raise ReadDataError(msg)

    def _is_statement_ready_to_fetch(self):
        status: StatementStatus = self.statement_api.get_statement(self.statement_id).status
        state: StatementState = status.state
        self.raise_for_status(status)

        if state in (StatementState.PENDING, StatementState.RUNNING):
            return False

        if state == StatementState.SUCCEEDED:
            return True

        raise ValueError(f'Unexpected statement state: {state}')

    def __iter__(self):
        self.wait()

        return self

    def _download_chunk(self, chunk_index, chunk_save_dir: Path = None) -> Union[bytes, Path]:
        chunk_description = self.statement_api.get_statement_result_chunk_n(
            statement_id=self.statement_id,
            chunk_index=chunk_index,
        )
        assert chunk_description.external_links, 'External links are not found in the chunk description'
        chunk = self._requests_session.get(chunk_description.external_links[0].external_link)
        chunk.raise_for_status()
        if not chunk_save_dir:
            return chunk.content

        save_path = chunk_save_dir / f'{chunk_index}.parquet'
        with open(save_path, 'wb') as f:
            f.write(arrow_stream2parquet(chunk.content))
        return save_path

    def __next__(self) -> dict:
        if self._chunk_ptr < self._total_chunk_count:
            if self._cached_batch and self._cached_batch_idx < len(self._cached_batch):
                record = self._cached_batch[self._cached_batch_idx]
                self._cached_batch_idx += 1
                return record
            else:
                chunk_bytes = self._download_chunk(chunk_index=self._chunk_ptr)
                self._cached_batch = pa.ipc.open_stream(chunk_bytes).read_pandas().to_dict('records')
                self._chunk_ptr += 1
                self._cached_batch_idx = 0
                return self.__next__()
        if self._cached_batch_idx < len(self._cached_batch):
            record = self._cached_batch[self._cached_batch_idx]
            self._cached_batch_idx += 1
            return record

        raise StopIteration

    def wait(self, timeout: int = STATEMENT_API_TIMEOUT) -> None:
        """
        Polls the statement status until it's ready to fetch the results.
        :param timeout: timeout in seconds
        """
        poll_start_time = time.time()
        while not self._is_statement_ready_to_fetch():
            logger.info(
                'Statement %s is not ready to fetch. Waiting for %s seconds',
                self.statement_id,
                SLEEP_TIME,
            )
            if time.time() - poll_start_time > timeout:
                raise ReadDataError(f'Statement {self.statement_id} is not ready to fetch. Timeout')

            time.sleep(SLEEP_TIME)

        logger.info('Statement %s is ready to fetch', self.statement_id)



from abc import ABC, abstractmethod
from functools import cache
from typing import Optional, Union

import attrs
from azure.identity import DefaultAzureCredential
from cachetools import TTLCache, cachedmethod

from dbxio.core.auth import (
    AZ_CRED_PROVIDER_TYPE,
    ClusterCredentials,
    SemiConfiguredClusterCredentials,
    check_is_airflow_installed,
    get_airflow_variables,
    get_local_variables,
)
from dbxio.core.exceptions import InsufficientCredentialsError, UnavailableAuthError
from dbxio.utils.databricks import ClusterType
from dbxio.utils.logging import get_logger

logger = get_logger()


class BaseAuthProvider(ABC):
    """
    Interface for the authentication provider.
    The authentication provider is responsible for fetching credentials to Databricks.
    """

    def __init__(self, **kwargs):
        raise NotImplementedError

    @property
    @abstractmethod
    def cluster_type(self) -> ClusterType:
        raise NotImplementedError

    @property
    @abstractmethod
    def az_cred_provider(self) -> AZ_CRED_PROVIDER_TYPE:
        raise NotImplementedError

    @property
    @abstractmethod
    def semi_configured_credentials(self) -> Optional[SemiConfiguredClusterCredentials]:
        raise NotImplementedError

    @abstractmethod
    def get_credentials(
        self,
        semi_configured_credentials: SemiConfiguredClusterCredentials = None,
    ) -> ClusterCredentials:
        raise NotImplementedError

    @classmethod
    def from_semi_configured_credentials(
        cls,
        http_path: str,
        server_hostname: str,
        cluster_type: ClusterType,
        az_cred_provider: AZ_CRED_PROVIDER_TYPE = None,
    ) -> 'BaseAuthProvider':
        semi_configured_credentials = cls.build_semi_configured_credentials(http_path, server_hostname)
        assert semi_configured_credentials is not None, 'http_path and server_hostname must be provided together'

        return cls(
            cluster_type=cluster_type,
            az_cred_provider=az_cred_provider or DefaultAzureCredential(),
            semi_configured_credentials=semi_configured_credentials,
        )

    @staticmethod
    def build_semi_configured_credentials(
        http_path: str = None,
        server_hostname: str = None,
    ) -> Union[SemiConfiguredClusterCredentials, None]:
        if http_path is None or server_hostname is None:
            return None

        return SemiConfiguredClusterCredentials(
            http_path=http_path,
            server_hostname=server_hostname,
        )


@attrs.define(slots=True)
class ClusterEnvAuthProvider(BaseAuthProvider):
    """
    Environment variables credential provider for Databricks.
    Assumes that either the environment variables are set or http path and server hostname are provided.
    """

    cluster_type: ClusterType = attrs.field(validator=attrs.validators.instance_of(ClusterType))
    az_cred_provider: AZ_CRED_PROVIDER_TYPE = attrs.field(factory=DefaultAzureCredential)

    semi_configured_credentials: Optional[SemiConfiguredClusterCredentials] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(SemiConfiguredClusterCredentials)),
    )

    _cache: TTLCache = attrs.Factory(lambda: TTLCache(maxsize=1024, ttl=60 * 15))

    @cachedmethod(lambda self: self._cache)
    def get_credentials(self) -> ClusterCredentials:
        try:
            return get_local_variables(self.az_cred_provider, self.semi_configured_credentials)
        except (TypeError, ValueError):
            logger.debug('ClusterEnvAuthProvider is not available. Not all environment variables are configured.')
            raise InsufficientCredentialsError


@attrs.define(slots=True)
class ClusterAirflowAuthProvider(BaseAuthProvider):
    """
    Credential provider for Databricks using Airflow variables.
    """

    cluster_type: ClusterType = attrs.field(validator=attrs.validators.instance_of(ClusterType))
    az_cred_provider: AZ_CRED_PROVIDER_TYPE = attrs.field(factory=DefaultAzureCredential)

    semi_configured_credentials: Optional[SemiConfiguredClusterCredentials] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(SemiConfiguredClusterCredentials)),
    )

    _cache: TTLCache = attrs.Factory(lambda: TTLCache(maxsize=1024, ttl=60 * 15))

    @cachedmethod(lambda self: self._cache)
    def get_credentials(
        self,
    ) -> ClusterCredentials:
        try:
            if not check_is_airflow_installed():
                logger.debug('ClusterAirflowAuthProvider is not available. Airflow is not installed.')
                raise InsufficientCredentialsError
            return get_airflow_variables(self.cluster_type, self.semi_configured_credentials)
        except TypeError:
            logger.debug('ClusterAirflowAuthProvider is not available. Some environment variables are not configured.')
            raise InsufficientCredentialsError
        except KeyError:
            logger.debug('ClusterAirflowAuthProvider is not available. Airflow variables are not configured.')
            raise InsufficientCredentialsError


class DefaultCredentialProvider:
    """
    Default credential provider for Databricks.
    It tries to get the credentials from all registered providers. If credentials are provided, it stops.
    """

    def __init__(
        self,
        cluster_type: ClusterType,
        az_cred_provider: AZ_CRED_PROVIDER_TYPE = None,
        http_path: str = None,
        server_hostname: str = None,
        lazy: bool = True,
    ):
        self.cluster_type = cluster_type
        self.az_cred_provider = az_cred_provider or DefaultAzureCredential()
        self.semi_configured_credentials = BaseAuthProvider.build_semi_configured_credentials(
            http_path, server_hostname
        )
        if http_path or server_hostname:
            assert (
                self.semi_configured_credentials is not None
            ), 'http_path and server_hostname must be provided together'

        self._chain = [ClusterAirflowAuthProvider, ClusterEnvAuthProvider]

        self._successful_provider: Union[BaseAuthProvider, None] = None
        if not lazy:
            assert (
                self.semi_configured_credentials is not None
            ), 'semi_configured_credentials must be provided if not lazy'
            self.ensure_set_auth_provider()

    def ensure_set_auth_provider(self) -> None:
        for provider_type in self._chain:
            try:
                provider = provider_type(
                    cluster_type=self.cluster_type,
                    az_cred_provider=self.az_cred_provider,
                    semi_configured_credentials=self.semi_configured_credentials,
                )
                provider.get_credentials()
                self._successful_provider = provider
                return
            except InsufficientCredentialsError:
                pass
        raise UnavailableAuthError('No available provider found')

    def get_credentials(self) -> ClusterCredentials:
        if self._successful_provider is None:
            self.ensure_set_auth_provider()

        assert self._successful_provider is not None
        return self._successful_provider.get_credentials()


@attrs.frozen(slots=True)
class BareAuthProvider(BaseAuthProvider):
    """
    Bare credentials provider for Databricks. This provider is used when the user provides the credentials directly.
    """

    access_token: str = attrs.field(validator=attrs.validators.instance_of(str))
    server_hostname: str = attrs.field(validator=attrs.validators.instance_of(str))
    http_path: str = attrs.field(validator=attrs.validators.instance_of(str))
    cluster_type: ClusterType = attrs.field(validator=attrs.validators.instance_of(ClusterType))
    az_cred_provider: AZ_CRED_PROVIDER_TYPE = attrs.field(factory=DefaultAzureCredential)

    semi_configured_credentials: None = attrs.field(default=None, init=False)

    @cache
    def get_credentials(self, **kwargs) -> ClusterCredentials:
        return ClusterCredentials(
            access_token=self.access_token,
            server_hostname=self.server_hostname,
            http_path=self.http_path,
        )



DATABRICKS_ACCESS_TOKEN = 'DATABRICKS_ACCESS_TOKEN'
DATABRICKS_SERVER_HOSTNAME = 'DATABRICKS_SERVER_HOSTNAME'
DATABRICKS_HTTP_PATH = 'DATABRICKS_HTTP_PATH'

DBX_FORCE_LOCAL = 'DBX_FORCE_LOCAL'

AIRFLOW_UNIQUE_NAME = 'AIRFLOW_UNIQUE_NAME'



import enum

from dbxio.core.cloud.aws import AWS_ClOUD_PROVIDER_NAME
from dbxio.core.cloud.azure import AZURE_ClOUD_PROVIDER_NAME
from dbxio.core.cloud.client.object_storage import ObjectStorageClient
from dbxio.core.cloud.gcp import GCP_ClOUD_PROVIDER_NAME


@enum.unique
class CloudProvider(str, enum.Enum):
    AWS = AWS_ClOUD_PROVIDER_NAME
    AZURE = AZURE_ClOUD_PROVIDER_NAME
    GCP = GCP_ClOUD_PROVIDER_NAME


__all__ = [
    'ObjectStorageClient',
    'CloudProvider',
]



import enum
from typing import Optional, Union

import attrs

from dbxio.delta.table_schema import TableSchema
from dbxio.sql.types import BaseType


class TableType(enum.Enum):
    Managed = 'Managed'
    Unmanaged = 'Unmanaged'


class TableFormat(enum.Enum):
    TEXT = 'TEXT'
    AVRO = 'AVRO'
    BINARYFILE = 'BINARYFILE'
    CSV = 'CSV'
    JSON = 'JSON'
    PARQUET = 'PARQUET'
    ORC = 'ORC'
    DELTA = 'DELTA'
    JDBC = 'JDBC'
    LIBSVM = 'LIBSVM'


class Materialization(enum.Enum):
    Table = 'Table'
    View = 'View'


@attrs.define
class TableAttributes:
    location: Optional[str] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )
    partitioned_by: Optional[list[str]] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(str),
                iterable_validator=attrs.validators.instance_of(list),
            )
        ),
    )
    properties: Optional[dict[str, str]] = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(dict)),
    )


def _table_schema_converter(schema: Union[dict[str, BaseType], list[dict[str, BaseType]], TableSchema]) -> TableSchema:
    return TableSchema.from_obj(schema)


def _table_identifier_validator(instance, attribute, value):
    if len(value.split('.')) != 3:
        raise ValueError(
            "It's not allowed to create/or use tables from default "
            f'database (got {value}). Please specify table path in the following format: '
            '<catalog>.<schema>.<table_name>'
        )


@attrs.define(slots=True)
class Table:
    """
    Represents a table in Databricks Unity Catalog. Used to define views as well as tables.
    """

    table_identifier: str = attrs.field(
        validator=[
            attrs.validators.instance_of(str),
            _table_identifier_validator,
        ]
    )
    table_format: TableFormat = attrs.field(
        default=TableFormat.DELTA,
        validator=attrs.validators.instance_of(TableFormat),
    )
    materialization: Materialization = attrs.field(
        default=Materialization.Table,
        validator=attrs.validators.instance_of(Materialization),
    )
    schema: Optional[Union[dict[str, BaseType], list[dict[str, BaseType]], TableSchema]] = attrs.field(
        default=None,
        converter=_table_schema_converter,
    )
    attributes: TableAttributes = attrs.field(factory=TableAttributes)

    @classmethod
    def from_obj(cls, obj: Union[str, 'Table']):
        if isinstance(obj, Table):
            return obj
        else:
            return Table(table_identifier=obj)

    @property
    def safe_table_identifier(self):
        """
        Returns table identifier with special characters replaced with underscores and wrapped in backticks.
        """
        trunc_ti = self.table_identifier.translate(
            str.maketrans('!"#$%&\'()*+,/:;<=>?@[\\]^`{|}~', '_____________________________')
        )
        return '.'.join([f'`{ti_part}`' for ti_part in trunc_ti.split('.')])

    @property
    def is_unmanaged(self) -> bool:
        return bool(self.attributes.location)

    @property
    def full_external_path(self) -> Optional[str]:
        return f'{self.table_format.name}.`{self.attributes.location}`' if self.is_unmanaged else None



import os
import time
from typing import Union

import attrs
from azure.core.credentials import TokenCredential
from azure.core.exceptions import ClientAuthenticationError
from azure.identity import AzureCliCredential, DefaultAzureCredential

from dbxio.utils.databricks import ClusterType
from dbxio.utils.env import (
    AIRFLOW_UNIQUE_NAME,
    DATABRICKS_ACCESS_TOKEN,
    DATABRICKS_HTTP_PATH,
    DATABRICKS_SERVER_HOSTNAME,
    DBX_FORCE_LOCAL,
)
from dbxio.utils.logging import get_logger

AZ_DBX_SCOPE = '2ff814a6-3304-4ab8-85cb-cd0e6f879c1d/.default'
DBX_DEFAULT_HOST = 'databricks_default_host'
DBX_DEFAULT_SQL_HTTP_PATH = 'databricks_default_sql_http_path'
DBX_DEFAULT_WAREHOUSE_SQL_HTTP_PATH = 'databricks_default_warehouse_sql_http_path'

AZ_CRED_PROVIDER_TYPE = Union[TokenCredential, DefaultAzureCredential, AzureCliCredential]

logger = get_logger()


@attrs.define(slots=True)
class ClusterCredentials:
    """
    Credentials for Databricks cluster.
    Is fully configured if all fields are set.

    Access token can be PAT or short-lived token.
    """

    access_token: str = attrs.field(validator=attrs.validators.instance_of(str))
    server_hostname: str = attrs.field(validator=attrs.validators.instance_of(str))
    http_path: str = attrs.field(validator=attrs.validators.instance_of(str))


@attrs.frozen(slots=True)
class SemiConfiguredClusterCredentials:
    """
    Credentials for Databricks cluster without access token.
    In this case, access token will be generated from a credential provider.
    """

    http_path: str = attrs.field(validator=attrs.validators.instance_of(str))
    server_hostname: str = attrs.field(validator=attrs.validators.instance_of(str))


def check_is_airflow_installed() -> bool:
    try:
        import airflow.exceptions  # noqa
        import airflow.models  # noqa
        from airflow.providers.databricks.hooks.databricks import DatabricksHook  # noqa
    except ModuleNotFoundError:
        return False
    return True


def _get_token_airflow(retries: int = 5) -> Union[str, None]:
    from airflow.exceptions import AirflowNotFoundException
    from airflow.models import Connection as AirflowConnection
    from airflow.providers.databricks.hooks.databricks import DatabricksHook

    for retry_i in range(retries):
        try:
            c = AirflowConnection.get_connection_from_secrets(DatabricksHook.default_conn_name)
            if dbx_token := c.get_password():
                return dbx_token
        except ModuleNotFoundError:
            logger.warning('Databricks provider is not installed, not retryable error')
            break
        except AirflowNotFoundException as e:
            logger.warning(f'Could not find databricks default connection: {e}')
            time.sleep(5 * 2**retry_i)

    return None


def _get_token_az(az_cred_provider: AZ_CRED_PROVIDER_TYPE = None, retries: int = 5) -> Union[str, None]:
    """
    Get Databricks short-lived token from Azure credential provider.
    """
    az_cred_provider = az_cred_provider or DefaultAzureCredential()
    for retry_i in range(retries):
        try:
            return az_cred_provider.get_token(AZ_DBX_SCOPE).token
        except ClientAuthenticationError as e:
            logger.warning(f'Could not get Databricks token due to connection error to Azure Vault: {e}')
            time.sleep(5 * 2**retry_i)
    return None


def get_token(az_cred_provider: AZ_CRED_PROVIDER_TYPE = None, retries: int = 5) -> str:
    """
    Get token from airflow environment or Azure credential provider.
    """
    assert retries >= 1, 'Retries must be a positive integer'

    force_local = os.getenv(DBX_FORCE_LOCAL)
    if not force_local and check_is_airflow_installed():
        if token_from_airflow := _get_token_airflow():
            return token_from_airflow

    if token_from_az := _get_token_az(az_cred_provider):
        return token_from_az

    raise RuntimeError('Could not get Databricks token')


def _get_airflow_http_path(cluster_type: ClusterType) -> str:
    """
    Get default HTTP path for Databricks cluster from Airflow variables.
    """
    from airflow.models import Variable

    airflow_unique_name = os.getenv(AIRFLOW_UNIQUE_NAME)

    if cluster_type == ClusterType.ALL_PURPOSE:
        return Variable.get(f'{airflow_unique_name}-{DBX_DEFAULT_SQL_HTTP_PATH}')
    elif cluster_type == ClusterType.SQL_WAREHOUSE:
        return Variable.get(f'{airflow_unique_name}-{DBX_DEFAULT_WAREHOUSE_SQL_HTTP_PATH}')
    else:
        raise ValueError(f'Unknown cluster type: {cluster_type}')


def _get_airflow_server_hostname() -> str:
    """
    Get default server hostname for Databricks cluster from Airflow variables.
    """
    from airflow.models import Variable

    airflow_unique_name = os.getenv(AIRFLOW_UNIQUE_NAME)
    return Variable.get(f'{airflow_unique_name}-{DBX_DEFAULT_HOST}')


def get_airflow_variables(
    cluster_type: ClusterType,
    semi_configured_credentials: SemiConfiguredClusterCredentials = None,
) -> ClusterCredentials:
    """
    Get credentials for Databricks cluster from Airflow variables.
    """
    http_path = (
        semi_configured_credentials.http_path
        if semi_configured_credentials is not None
        else _get_airflow_http_path(cluster_type)
    )
    server_hostname = (
        semi_configured_credentials.server_hostname
        if semi_configured_credentials is not None
        else _get_airflow_server_hostname()
    )

    return ClusterCredentials(
        access_token=get_token(),
        server_hostname=server_hostname,
        http_path=http_path,
    )


def get_local_variables(
    az_cred_provider: AZ_CRED_PROVIDER_TYPE = None,
    semi_configured_credentials: SemiConfiguredClusterCredentials = None,
) -> ClusterCredentials:
    """
    Get credentials for Databricks cluster from local environment variables.
    Access token can be generated by Azure credential provider.
    """
    access_token = os.getenv(DATABRICKS_ACCESS_TOKEN)
    if not access_token:
        logger.debug('No DATABRICKS_ACCESS_TOKEN found, trying to get token from Azure')
        access_token = get_token(az_cred_provider)
    http_path = (
        semi_configured_credentials.http_path
        if semi_configured_credentials is not None
        else os.getenv(DATABRICKS_HTTP_PATH)
    )
    if not http_path:
        raise ValueError('DATABRICKS_HTTP_PATH is not set')
    server_hostname = (
        semi_configured_credentials.server_hostname
        if semi_configured_credentials is not None
        else os.getenv(DATABRICKS_SERVER_HOSTNAME)
    )
    if not server_hostname:
        raise ValueError('DATABRICKS_SERVER_HOSTNAME is not set')

    return ClusterCredentials(
        access_token=access_token,
        server_hostname=server_hostname,
        http_path=http_path,
    )



import requests
from requests.adapters import HTTPAdapter, Retry


def get_session(tries: int = 3, backoff: float = 0.1) -> requests.Session:
    session = requests.Session()
    retries = Retry(total=tries, backoff_factor=backoff, status_forcelist=[429, 500, 502, 503, 504])
    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.mount('https://', HTTPAdapter(max_retries=retries))
    return session



from dbxio.core.auth import ClusterCredentials, SemiConfiguredClusterCredentials, get_token
from dbxio.core.client import DbxIOClient, DefaultDbxIOClient, DefaultSqlDbxIOClient
from dbxio.core.credentials import (
    BareAuthProvider,
    BaseAuthProvider,
    ClusterAirflowAuthProvider,
    ClusterEnvAuthProvider,
    DefaultCredentialProvider,
)
from dbxio.core.exceptions import DbxIOTypeError, InsufficientCredentialsError, ReadDataError, UnavailableAuthError
from dbxio.core.settings import CloudProvider, Settings

__all__ = [
    'get_token',
    'ClusterCredentials',
    'SemiConfiguredClusterCredentials',
    'DbxIOClient',
    'DefaultDbxIOClient',
    'DefaultSqlDbxIOClient',
    'BaseAuthProvider',
    'ClusterEnvAuthProvider',
    'ClusterAirflowAuthProvider',
    'DefaultCredentialProvider',
    'BareAuthProvider',
    'DbxIOTypeError',
    'InsufficientCredentialsError',
    'UnavailableAuthError',
    'ReadDataError',
    'Settings',
    'CloudProvider',
]



import os
import uuid
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Union

import attrs
from databricks.sdk.service.catalog import VolumeType
from tenacity import retry, stop_after_attempt, wait_fixed

from dbxio.blobs.block_upload import upload_file
from dbxio.blobs.download import download_blob_tree
from dbxio.core.cloud.client.object_storage import ObjectStorageClient
from dbxio.sql.results import _FutureBaseResult
from dbxio.utils.blobs import blobs_registries
from dbxio.utils.logging import get_logger

if TYPE_CHECKING:
    from dbxio.core.client import DbxIOClient

DEFAULT_EXTERNAL_LOCATION = 'default_external_location'

logger = get_logger()


@attrs.define(slots=True)
class Volume:
    """
    Represents a volume in Databricks.

    Storage_location is required for external volumes.
    It must satisfy the following pattern:
        abfss://<container_name>@<storage_name>.dfs.core.windows.net/<blobs_path>}
    """

    catalog: str = attrs.field(validator=attrs.validators.instance_of(str))
    schema: str = attrs.field(validator=attrs.validators.instance_of(str))
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    path: str = attrs.field(default='', validator=attrs.validators.instance_of(str))
    volume_type: VolumeType = attrs.field(
        default=VolumeType.MANAGED,
        validator=attrs.validators.instance_of(VolumeType),
    )
    storage_location: Union[str, None] = None

    def __attrs_post_init__(self):
        if self.volume_type == VolumeType.EXTERNAL and not self.storage_location:
            raise ValueError('storage_location is required for external volumes')

    def __truediv__(self, path: str):
        return Volume(
            catalog=self.catalog,
            schema=self.schema,
            name=self.name,
            path=str(Path(self.path) / Path(path)),
            volume_type=self.volume_type,
            storage_location=self.storage_location,
        )

    @property
    def full_name(self):
        return f'{self.catalog}.{self.schema}.{self.name}'

    @property
    def safe_full_name(self):
        return f'`{self.catalog}`.`{self.schema}`.`{self.name}`'

    @property
    def mount_path(self):
        return str(Path(f'/Volumes/{self.catalog}/{self.schema}/{self.name}') / self.path)

    def is_managed(self):
        return self.volume_type is VolumeType.MANAGED

    def is_external(self):
        return self.volume_type is VolumeType.EXTERNAL


def create_volume(volume: Volume, client: 'DbxIOClient') -> None:
    client.workspace_api.volumes.create(
        catalog_name=volume.catalog,
        schema_name=volume.schema,
        name=volume.name,
        volume_type=volume.volume_type,
        storage_location=volume.storage_location,
    )
    logger.info(f'Volume {volume.safe_full_name} was successfully created.')


def _exists_volume(volume: Volume, client: 'DbxIOClient') -> bool:
    for v in client.workspace_api.volumes.list(catalog_name=volume.catalog, schema_name=volume.schema):
        if v.name == volume.name:
            return True

    return False


def _download_external_volume(local_path: Path, storage_location: str):
    object_storage = ObjectStorageClient.from_url(storage_location)
    download_blob_tree(
        object_storage_client=object_storage,
        local_path=local_path,
        prefix_path=object_storage.blobs_path,
    )


@retry(stop=stop_after_attempt(3), wait=wait_fixed(10))
def _download_single_file_from_managed_volume(local_path: Path, file_path: str, client: 'DbxIOClient'):
    with open(local_path / Path(file_path).name, 'wb') as f:
        response_content = client.workspace_api.files.download(file_path).contents
        if response_content:
            f.write(response_content.read())
        else:
            raise ValueError(f'Failed to download file {file_path}, got None')


def _download_managed_volume(local_path: Path, volume: Volume, client: 'DbxIOClient'):
    for file in client.workspace_api.files.list_directory_contents(volume.mount_path):
        if file.name is None or file.path is None:
            raise ValueError(f'File {file} has no name or path')

        if file.is_directory:
            (local_path / file.name).mkdir(parents=True, exist_ok=True)
            _download_managed_volume(local_path / file.name, volume / file.name, client)
        else:
            _download_single_file_from_managed_volume(local_path, file.path, client)


def download_volume(
    path: Union[str, Path],
    catalog_name: str,
    schema_name: str,
    volume_name: str,
    client: 'DbxIOClient',
) -> Path:
    """
    Downloads files from a volume in Databricks to a local directory.
    Volume must be an external volume.

    :param path: local directory to download files to. Must be a directory.
    :param catalog_name: databricks catalog name
    :param schema_name: databricks schema name
    :param volume_name: volume name
    :param client: dbxio client
    """
    path = Path(path)
    if not path.exists() or not path.is_dir():
        raise ValueError('Path must be an existing directory')
    volume_info = client.workspace_api.volumes.read(f'{catalog_name}.{schema_name}.{volume_name}')
    if volume_info.volume_type is None:
        raise ValueError(f'Volume {catalog_name}.{schema_name}.{volume_name} does not exist')

    volume = Volume(
        catalog=catalog_name,
        schema=schema_name,
        name=volume_name,
        volume_type=volume_info.volume_type,
        storage_location=volume_info.storage_location,
    )
    logger.info(f'Downloading volume: {volume.full_name}')
    logger.debug(f'Volume type: {volume.volume_type}')

    if volume_info.volume_type == VolumeType.EXTERNAL:
        assert volume_info.storage_location, f'External volume must have a storage location, got {volume_info=}'
        _download_external_volume(local_path=path, storage_location=volume_info.storage_location)
    elif volume_info.volume_type == VolumeType.MANAGED:
        _download_managed_volume(local_path=path, volume=volume, client=client)
    else:
        raise ValueError(f'Unknown volume type: {volume_info.volume_type}')

    logger.info(f'Volume {volume.full_name} was successfully downloaded to {path}')

    return path


def write_volume(
    path: Union[str, Path],
    catalog_name: str,
    schema_name: str,
    volume_name: str,
    client: 'DbxIOClient',
    create_volume_if_not_exists: bool = True,
    max_concurrency: int = 1,
    force: bool = False,
) -> None:
    """
    writes the contents of a local directory to a volume in Databricks.

    :param path: local directory or file to upload
    :param catalog_name: databricks catalog name
    :param schema_name: databricks schema name
    :param volume_name: volume name
    :param client: dbxio client
    :param create_volume_if_not_exists: flag to create the volume if it does not exist
    :param max_concurrency: the maximum number of threads to use for uploading files
    :param force: a flag to break leases if the blob is already in use
    """
    path = Path(path)

    catalog = client.workspace_api.catalogs.get(catalog_name)
    catalog_properties = catalog.properties if catalog.properties else {}

    default_external_location = catalog_properties.get(DEFAULT_EXTERNAL_LOCATION)
    if not default_external_location:
        raise ValueError(
            'Catalog does not have a default external location (aka container in storage account). '
            'Ask data engineers for creation of a default external location for your catalog.'
        )
    storage_url = client.workspace_api.external_locations.get(default_external_location).url
    assert storage_url, f'External location {default_external_location} does not have a URL'
    object_storage_client = ObjectStorageClient.from_url(storage_url)

    operation_uuid = str(uuid.uuid4())
    with blobs_registries(object_storage_client, keep_blobs=True) as (blobs, metablobs):
        # here all files in the path, including subdirectories, are uploaded to the blob storage.
        # only "hidden" files (those starting with a dot) are skipped
        for file in path.glob('**/*'):
            if file.is_file() and not file.name.startswith('.'):
                upload_file(
                    path=file,
                    local_path=path,
                    object_storage_client=object_storage_client,
                    operation_uuid=operation_uuid,
                    blobs=blobs,
                    metablobs=metablobs,
                    max_concurrency=max_concurrency,
                    force=force,
                )
        object_storage_client.blobs_path = str(os.path.commonpath(blobs))
        storage_location = object_storage_client.to_url()
    volume = Volume(
        catalog=catalog_name,
        schema=schema_name,
        name=volume_name,
        volume_type=VolumeType.EXTERNAL,
        storage_location=storage_location,
    )

    if not _exists_volume(volume, client):
        if create_volume_if_not_exists:
            create_volume(volume=volume, client=client)
        else:
            raise ValueError(
                f'Volume {volume_name} does not exist in schema {schema_name} of catalog '
                f'{catalog_name} and create_volume_if_not_exists is False.'
            )


def set_tags_on_volume(volume: Volume, tags: dict[str, str], client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Sets tags on a volume.
    Each tag is a key-value pair of strings.
    """
    if not tags:
        raise ValueError('tags must be a non-empty dictionary')

    set_tags_query = dedent(f"""
    ALTER VOLUME {volume.safe_full_name}
    SET TAGS ({', '.join([f'"{k}" = "{v}"' for k, v in tags.items()])})
    """)

    return client.sql(set_tags_query)


def unset_tags_on_volume(volume: Volume, tags: list[str], client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Unsets tags on a volume.
    """
    if not tags:
        raise ValueError('tags must be a non-empty list')

    unset_tags_query = dedent(f"""
    ALTER VOLUME {volume.safe_full_name}
    UNSET TAGS ({', '.join([f'"{tag}"' for tag in tags])})
    """)

    return client.sql(unset_tags_query)


def get_tags_on_volume(volume: Volume, client: 'DbxIOClient') -> dict[str, str]:
    """
    Returns the tags on a volume.
    """
    information_schema_query = dedent(f"""
    select tag_name, tag_value
    from system.information_schema.volume_tags
    where
        catalog_name = '{volume.catalog}'
        and schema_name = '{volume.schema}'
        and volume_name = '{volume.name}'
    """)

    tags = {}
    with client.sql(information_schema_query) as result:
        for row in result:
            tags[row['tag_name']] = row['tag_value']

    return tags


def set_comment_on_volume(
    volume: Volume,
    comment: Union[str, None],
    client: 'DbxIOClient',
) -> _FutureBaseResult:
    """
    Sets a comment on a volume.
    Description comment supports Markdown.

    If the comment is None, the comment will be removed.
    """
    set_comment_query = dedent(
        f"""COMMENT ON VOLUME {volume.safe_full_name} IS {f'"{comment}"' if comment else 'NULL'}"""
    )

    return client.sql(set_comment_query)


def unset_comment_on_volume(volume: Volume, client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Unsets the comment on a volume.
    """
    return set_comment_on_volume(volume=volume, comment=None, client=client)


def get_comment_on_volume(volume: Volume, client: 'DbxIOClient') -> Union[str, None]:
    """
    Returns the comment on a volume.
    """
    information_schema_query = dedent(f"""
    select comment
    from system.information_schema.volumes
    where
        catalog_name = '{volume.catalog}'
        and schema_name = '{volume.schema}'
        and volume_name = '{volume.name}'
    """)

    with client.sql(information_schema_query) as result:
        for row in result:
            return row['comment']

    return None



from functools import cached_property
from pathlib import Path
from typing import Any, Dict, Optional, Union

import attrs
from databricks.sdk import StatementExecutionAPI, WorkspaceClient

from dbxio.core.credentials import AZ_CRED_PROVIDER_TYPE, BaseAuthProvider, DefaultCredentialProvider
from dbxio.core.settings import Settings
from dbxio.sql.query import BaseDatabricksQuery
from dbxio.sql.results import _FutureBaseResult
from dbxio.sql.sql_driver import SQLDriver, get_sql_driver
from dbxio.utils.databricks import ClusterType


@attrs.define(slots=True)
class DbxIOClient:
    """
    Client for interacting with Databricks.

    To create a client, you can use one of the following methods:
    >>> client = DbxIOClient.from_cluster_settings(
    ...     http_path='<YOUR_HTTP_PATH>',
    ...     server_hostname='<YOUR_SERVER_HOSTNAME>',
    ...     cluster_type=ClusterType.ALL_PURPOSE,
    ... )
    or
    >>> client = DbxIOClient.from_auth_provider(
    ...     auth_provider=DefaultCredentialProvider(
    ...         cluster_type=ClusterType.ALL_PURPOSE,
    ...         az_cred_provider=AzureCliCredential(),
    ...         http_path='<YOUR_HTTP_PATH>',
    ...         server_hostname='<YOUR_SERVER_HOSTNAME>',
    ...     ),
    ... )
    """

    credential_provider: Union[BaseAuthProvider, DefaultCredentialProvider] = attrs.field(
        validator=attrs.validators.instance_of((BaseAuthProvider, DefaultCredentialProvider))
    )
    settings: Settings = attrs.field(validator=attrs.validators.instance_of(Settings))

    session_configuration: Optional[Dict[str, Any]] = None

    @classmethod
    def from_cluster_settings(
        cls,
        http_path: str,
        server_hostname: str,
        cluster_type: ClusterType,
        az_cred_provider: AZ_CRED_PROVIDER_TYPE = None,
        settings: Optional[Settings] = None,
        **kwargs,
    ):
        """
        Create a client from the cluster settings. Use this method if you want to use PAT for authentication.
        """
        provider = DefaultCredentialProvider(
            cluster_type=cluster_type,
            az_cred_provider=az_cred_provider,
            http_path=http_path,
            server_hostname=server_hostname,
        )
        return cls.from_auth_provider(auth_provider=provider, settings=settings or Settings(), **kwargs)

    @classmethod
    def from_auth_provider(
        cls,
        auth_provider: Union[BaseAuthProvider, DefaultCredentialProvider],
        settings: Optional[Settings] = None,
        **kwargs,
    ):
        """
        Create a client from the auth provider.
        Use this method if you want to generate a short-lived token based on the available authentication method.
        """
        return cls(credential_provider=auth_provider, settings=settings or Settings(), **kwargs)

    @property
    def _cluster_credentials(self):
        return self.credential_provider.get_credentials()

    @property
    def workspace_api(self) -> WorkspaceClient:
        return WorkspaceClient(
            host=self._cluster_credentials.server_hostname,
            token=self._cluster_credentials.access_token,
        )

    @property
    def statement_api(self) -> StatementExecutionAPI:
        return self.workspace_api.statement_execution

    @cached_property
    def _sql_driver(self) -> SQLDriver:
        return get_sql_driver(
            cluster_type=self.credential_provider.cluster_type,
            cluster_credentials=self._cluster_credentials,
            statement_api=self.statement_api,
            session_configuration=self.session_configuration,
        )

    def sql(self, query: Union[str, BaseDatabricksQuery]) -> _FutureBaseResult:
        """
        Execute the SQL query and return the results as a future.
        The future can be waited with .wait() method or iterated over.

        Results might be combined into chunks, so it's recommended to flatten them if you need a list of rows.
        """
        return self._sql_driver.sql(query)

    def sql_to_files(
        self,
        query: Union[str, BaseDatabricksQuery],
        results_path: str,
        max_concurrency: int = 1,
    ) -> Path:
        """
        Execute the SQL query and save the results to the specified directory.
        Returns the path to the directory with the results including the statement ID.
        """

        return self._sql_driver.sql_to_files(query, results_path, max_concurrency)


class DefaultDbxIOClient(DbxIOClient):
    """
    Default client for all-purpose clusters.

    Short and user-friendly way to create a client if all required environment variables are set.
    """

    def __init__(self, session_configuration: Optional[Dict[str, Any]] = None):
        super().__init__(
            credential_provider=DefaultCredentialProvider(cluster_type=ClusterType.ALL_PURPOSE),
            session_configuration=session_configuration,
            settings=Settings(),
        )


class DefaultSqlDbxIOClient(DbxIOClient):
    """
    Default client for SQL warehouses.

    Short and user-friendly way to create a client if all required environment variables are set.
    """

    def __init__(self, session_configuration: Optional[Dict[str, Any]] = None):
        super().__init__(
            credential_provider=DefaultCredentialProvider(cluster_type=ClusterType.SQL_WAREHOUSE),
            session_configuration=session_configuration,
            settings=Settings(),
        )



from dbxio.utils.blobs import blobs_gc, blobs_registries
from dbxio.utils.databricks import ClusterType
from dbxio.utils.env import (
    AIRFLOW_UNIQUE_NAME,
    DATABRICKS_ACCESS_TOKEN,
    DATABRICKS_HTTP_PATH,
    DATABRICKS_SERVER_HOSTNAME,
    DBX_FORCE_LOCAL,
)
from dbxio.utils.http import get_session
from dbxio.utils.logging import get_logger

__all__ = [
    'ClusterType',
    'DATABRICKS_HTTP_PATH',
    'DATABRICKS_ACCESS_TOKEN',
    'DATABRICKS_SERVER_HOSTNAME',
    'DBX_FORCE_LOCAL',
    'AIRFLOW_UNIQUE_NAME',
    'blobs_registries',
    'blobs_gc',
    'get_session',
    'get_logger',
]



from typing import Any, Dict

from dbxio.delta.table_schema import TableSchema
from dbxio.sql.types import as_dbxio_type


def infer_schema(record: Dict[str, Any]) -> TableSchema:
    """
    Infer schema from a record. Optimal data types will be used.
    """
    schema = []
    for key, value in record.items():
        col = {'name': key, 'type': as_dbxio_type(value)}
        schema.append(col)
    return TableSchema.from_obj(schema)



from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Optional, Union

import attrs
from databricks import sql
from databricks.sdk.service.sql import (
    Disposition,
    ExecuteStatementRequestOnWaitTimeout,
    Format,
    StatementExecutionAPI,
    StatementParameterListItem,
)
from databricks.sql.client import Connection, Cursor

from dbxio.core.auth import ClusterCredentials
from dbxio.sql.query import QUERY_PARAMS_TYPE, BaseDatabricksQuery, ConstDatabricksQuery, log_query
from dbxio.sql.results import _FutureBaseResult, _FutureODBCResult, _FutureStatementApiResult
from dbxio.utils.databricks import ClusterType
from dbxio.utils.logging import get_logger

logger = get_logger()


@attrs.define(slots=True)
class SQLDriver:
    """
    Interface for SQL drivers.
    """

    @property
    @abstractmethod
    def cluster_type(self) -> ClusterType:
        raise NotImplementedError

    @property
    @abstractmethod
    def cluster_credentials(self) -> ClusterCredentials:
        raise NotImplementedError

    @abstractmethod
    def _sql_impl(self, built_query: str, built_params: QUERY_PARAMS_TYPE) -> _FutureBaseResult:
        raise NotImplementedError

    @abstractmethod
    def _sql_to_files_impl(
        self,
        built_query: str,
        built_params: QUERY_PARAMS_TYPE,
        results_path: str,
        max_concurrency: int = 1,
    ) -> Path:
        raise NotImplementedError

    def _build_query(self, query: Union[str, BaseDatabricksQuery]) -> tuple[str, QUERY_PARAMS_TYPE]:
        if isinstance(query, str):
            query = ConstDatabricksQuery(query)
        built_query, built_params = query.build(self.cluster_type)
        log_query(built_query)

        return built_query, built_params

    def sql(self, query: Union[str, BaseDatabricksQuery]) -> _FutureBaseResult:
        """
        Execute the SQL query and return the future with the results.
        """
        built_query, built_params = self._build_query(query)

        return self._sql_impl(built_query=built_query, built_params=built_params)

    def sql_to_files(
        self,
        query: Union[str, BaseDatabricksQuery],
        results_path: str,
        max_concurrency: int = 1,
    ) -> Path:
        """
        Execute the SQL query and save the results to the specified directory.
        Returns the path to the directory with the results including the statement ID.
        If statement ID is not provided by Databricks, it will be generated.
        """
        assert max_concurrency > 0, 'max_concurrency must be greater than 0'
        if results_path:
            assert Path(results_path).is_dir(), f'{results_path} is not a valid directory'

        built_query, built_params = self._build_query(query)

        return self._sql_to_files_impl(
            built_query=built_query,
            built_params=built_params,
            results_path=results_path,
            max_concurrency=max_concurrency,
        )


@attrs.define(slots=True)
class ODBCDriver(SQLDriver):
    """
    SQL driver for all-purpose clusters.

    Warning: ODBC connection can be extremely slow.
    Consider using the SQL Warehouse cluster type instead.

    Docs: https://docs.databricks.com/en/dev-tools/python-sql-connector.html
    """

    cluster_type: ClusterType = attrs.field(validator=attrs.validators.instance_of(ClusterType))
    cluster_credentials: ClusterCredentials = attrs.field(validator=attrs.validators.instance_of(ClusterCredentials))
    session_configuration: Optional[dict[str, Any]] = attrs.field(default=None)

    conn: Optional[Connection] = attrs.field(default=None, init=False, repr=False)
    cursor: Optional[Cursor] = attrs.field(default=None, init=False, repr=False)
    thread_pool: ThreadPoolExecutor = attrs.Factory(ThreadPoolExecutor)

    def as_dict(self) -> dict[str, Any]:
        return {
            'server_hostname': self.cluster_credentials.server_hostname,
            'http_path': self.cluster_credentials.http_path,
            'access_token': self.cluster_credentials.access_token,
            'session_configuration': self.session_configuration,
        }

    def _sql_impl(self, built_query: str, built_params: QUERY_PARAMS_TYPE) -> _FutureODBCResult:
        logger.warning(
            'You are using the ODBC connection to execute the query, because the cluster type is ALL_PURPOSE. '
            'ODBC connection is very slow. Consider using the SQL Warehouse cluster type instead.',
        )

        def _execute_sql_query():
            if not self.conn or not self.cursor or not self.cursor.open or not self.conn.open:
                self.conn = sql.connect(**self.as_dict())
                self.cursor = self.conn.cursor()

            self.cursor.execute(operation=built_query, parameters=built_params)
            return self.conn, self.cursor

        return _FutureODBCResult(self.thread_pool.submit(_execute_sql_query))

    def _sql_to_files_impl(
        self,
        built_query: str,
        built_params: QUERY_PARAMS_TYPE,
        results_path: str,
        max_concurrency: int = 1,
    ) -> Path:
        if max_concurrency > 1:
            logger.warning('Concurrent downloads of SQL results are not supported for ALL_PURPOSE clusters.')

        future = self._sql_impl(built_query, built_params)
        return future.download_and_save(results_path=results_path)


@attrs.define(slots=True)
class StatementAPIDriver(SQLDriver):
    """
    SQL driver for SQL warehouses.

    Docs: https://docs.databricks.com/en/dev-tools/sql-execution-tutorial.html
    """

    cluster_type: ClusterType = attrs.field(validator=attrs.validators.instance_of(ClusterType))
    cluster_credentials: ClusterCredentials = attrs.field(validator=attrs.validators.instance_of(ClusterCredentials))
    statement_api: StatementExecutionAPI = attrs.field(validator=attrs.validators.instance_of(StatementExecutionAPI))

    def _sql_impl(self, built_query: str, built_params: QUERY_PARAMS_TYPE) -> _FutureStatementApiResult:
        warehouse_id = self.cluster_credentials.http_path.split('/')[-1]
        assert warehouse_id, f'Warehouse ID is not found in the HTTP path, got {self.cluster_credentials.http_path}'
        if built_params:
            assert all(
                isinstance(p, StatementParameterListItem) for p in built_params
            ), f'Invalid parameters types, got {built_params}'

        statement_response = self.statement_api.execute_statement(
            statement=built_query,
            parameters=built_params,  # type: ignore
            warehouse_id=warehouse_id,
            disposition=Disposition.EXTERNAL_LINKS,
            format=Format.ARROW_STREAM,
            on_wait_timeout=ExecuteStatementRequestOnWaitTimeout.CONTINUE,
        )
        return _FutureStatementApiResult(statement_response, statement_api=self.statement_api)

    def _sql_to_files_impl(
        self,
        built_query: str,
        built_params: QUERY_PARAMS_TYPE,
        results_path: str,
        max_concurrency: int = 1,
    ) -> Path:
        result = self._sql_impl(built_query, built_params)
        return result.download_and_save(results_path=results_path, max_concurrency=max_concurrency)


def get_sql_driver(
    cluster_type: ClusterType,
    cluster_credentials: ClusterCredentials,
    statement_api: StatementExecutionAPI,
    session_configuration: Optional[dict[str, Any]] = None,
) -> SQLDriver:
    """
    Factory function to create the SQL driver based on the cluster type.
    """
    if cluster_type is ClusterType.ALL_PURPOSE:
        return ODBCDriver(
            cluster_type=cluster_type,
            cluster_credentials=cluster_credentials,
            session_configuration=session_configuration,
        )
    elif cluster_type is ClusterType.SQL_WAREHOUSE:
        return StatementAPIDriver(
            cluster_type=cluster_type,
            cluster_credentials=cluster_credentials,
            statement_api=statement_api,
        )
    else:
        raise ValueError(f'Unsupported cluster type: {cluster_type}')



import datetime
import json
import math
from abc import ABCMeta, abstractmethod
from decimal import Decimal
from typing import Any

import dateutil.parser
import numpy as np
import pyarrow as pa
from decorator import decorator

from dbxio.core.exceptions import DbxIOTypeError
from dbxio.utils.logging import get_logger

logger = get_logger()


@decorator
def nullable(function, *args):
    if args[1] is None:
        return NullType().serialize(args[1])
    if isinstance(args[1], str) and args[1] == 'NULL':
        return NullType().deserialize(args[1])
    return function(*args)


class BaseType(metaclass=ABCMeta):
    def __str__(self):
        """
        Each class name ends with `Type` suffix. This method returns the name without it in the upper case.
        """
        return self.__class__.__name__[:-4].upper()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseType):
            return NotImplemented
        return isinstance(self, other.__class__) and str(self) == str(other)

    def __instancecheck__(self, instance):
        return self.__eq__(instance)

    @abstractmethod
    def fit(self, obj) -> bool:
        raise NotImplementedError

    @abstractmethod
    def serialize(self, obj, unsafe: bool = False):
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, obj):
        raise NotImplementedError

    def raise_uncast_exception(self, obj):
        raise DbxIOTypeError(f'Passed uncastable value. Cannot cast `{obj}` to type {self}')

    def try_fit(self, obj, unsafe: bool = False):
        if not unsafe and not self.fit(obj):
            self.raise_uncast_exception(obj)


class TinyIntType(BaseType):
    """
    Represents 1-byte signed integer numbers.
    The range of numbers is from -128 to 127.

    Same as numpy.int8
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, np.integer):
            obj = int(obj)
        if not isinstance(obj, int):
            return False
        if -(2**7) <= int(obj) <= 2**7 - 1:
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'{int(obj)}Y'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('Y'):
            return int(obj.strip('Y'))
        return int(obj)


class SmallIntType(BaseType):
    """
    Represents 2-byte signed integer numbers.
    The range of numbers is from -32,768 to 32,767.

    Same as numpy.int16
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, np.integer):
            obj = int(obj)
        if not isinstance(obj, int):
            return False
        if -(2**15) <= obj <= 2**15 - 1:
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'{int(obj)}S'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('S'):
            return int(obj.strip('S'))
        return int(obj)


class IntType(BaseType):
    """
    Represents 4-byte signed integer numbers.
    The range of numbers is from -2,147,483,648 to 2,147,483,647.

    Same as numpy.int32
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, np.integer):
            obj = int(obj)
        if isinstance(obj, int) and -(2**31) <= obj <= 2**31 - 1:
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'{int(obj)}'

    @nullable
    def deserialize(self, obj):
        return int(obj)


class BigIntType(BaseType):
    """
    Represents 8-byte signed integer numbers.
    The range of numbers is from -9,223,372,036,854,775,808 to 9,223,372,036,854,775,807.

    Same as numpy.int64
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, np.integer):
            obj = int(obj)
        if isinstance(obj, int) and -(2**63) <= obj <= 2**63 - 1:
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'{int(obj)}L'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('L'):
            return int(obj.strip('L'))
        return int(obj)


class DecimalType(BaseType):
    """
    Represents numbers with a specified maximum precision and fixed scale.
    """

    def __init__(self, precision: int = 10, scale: int = 0):
        assert 1 <= precision <= 38, 'Precision must be between 1 and 38'
        assert 0 <= scale <= precision, 'Scale must be between 0 and precision'

        self.precision = precision
        self.scale = scale

    def __str__(self):
        return f'DECIMAL({self.precision},{self.scale})'

    def fit(self, obj) -> bool:
        if isinstance(obj, (int, float, Decimal)):
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'{str(obj)}BD'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('BD'):
            return float(obj.strip('BD'))
        return float(obj)


class FloatType(BaseType):
    """
    Represents 8-byte double-precision floating point numbers.
    The range of numbers is:
     > Negative infinity
     > -3.402E+38 to -1.175E-37
     > 0
     > +1.175E-37 to +3.402E+38
     > Positive infinity
     > NaN (not a number)

    Same as numpy.float32
    """

    def fit(self, obj) -> bool:
        if not isinstance(obj, (int, float, np.floating)):
            return False
        if isinstance(obj, np.floating):
            obj = float(obj)
        if obj in (float('inf'), -float('inf')):
            return True
        if -3.402e38 <= obj <= -1.175e-37:
            return True
        if +1.175e-37 <= obj <= +3.402e38:
            return True
        if obj == 0.0 or obj is None:
            return True
        if math.isnan(obj):
            return True

        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        if obj == float('inf'):
            return "float('inf')"
        if obj == -float('inf'):
            return "float('-inf')"
        return f'{str(obj)}F'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('F'):
            return float(obj.strip('F'))
        if isinstance(obj, str) and 'inf' in obj:
            if obj == "float('inf')":
                return float('inf')
            if obj == "float('-inf')":
                return -float('inf')
        return float(obj)


class DoubleType(BaseType):
    """
    Represents 8-byte double-precision floating point numbers.
    The range of numbers is:
     > Negative infinity
     > -1.79769E+308 to -2.225E-307
     > 0
     > +2.225E-307 to +1.79769E+308
     > Positive infinity
     > NaN (not a number)

    Same as numpy.float64
    """

    def fit(self, obj) -> bool:
        if not isinstance(obj, (int, float, np.floating)):
            return False
        if isinstance(obj, np.floating):
            obj = float(obj)
        if obj in (float('inf'), -float('inf')):
            return True
        if -1.79769e308 <= obj <= -2.225e-307:
            return True
        if +2.225e-307 <= obj <= +1.79769e308:
            return True
        if obj == 0.0 or obj is None:
            return True
        if math.isnan(obj):
            return True

        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        if obj == float('inf'):
            return "double('inf')"
        if obj == -float('inf'):
            return "double('-inf')"
        return f'{str(obj)}D'

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str) and obj.endswith('D'):
            return float(obj.strip('D'))
        if isinstance(obj, str) and 'inf' in obj:
            if obj == "double('inf')":
                return float('inf')
            if obj == "double('-inf')":
                return -float('inf')
        return float(obj)


class DateType(BaseType):
    """
    Represents values comprising values of fields year, month, and day, without a time-zone.
    The range of dates supported is June 23 -5877641 CE to July 11 +5881580 CE.
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, datetime.date) and not isinstance(obj, datetime.datetime):
            return True
        elif isinstance(obj, str):
            try:
                dtm = datetime.datetime.strptime(obj, '%Y-%m-%d')
                if dtm.time() == datetime.time(0, 0):
                    return True
                else:
                    return False
            except ValueError:
                return False
        else:
            return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f"DATE'{obj}'" if isinstance(obj, str) else f"DATE'{obj.isoformat()}'"

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, datetime.date):
            return obj
        if isinstance(obj, str) and obj.startswith("DATE'"):
            return datetime.date.fromisoformat(obj.strip("DATE''"))
        return datetime.datetime.strptime(obj, '%Y-%m-%d')


class TimestampType(BaseType):
    """
    Represents values comprising values of fields year, month, day, hour, minute, and second,
    with the session local time-zone. The timestamp value represents an absolute point in time.
    The range of timestamps supported is June 23 -5877641 CE to July 11 +5881580 CE.
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return True
        elif isinstance(obj, str):
            try:
                dateutil.parser.parse(obj)
                return True
            except dateutil.parser.ParserError:
                return False
        else:
            return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f"TIMESTAMP'{obj}'" if isinstance(obj, str) else f"TIMESTAMP'{obj.isoformat()}'"

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj
        if isinstance(obj, str) and obj.startswith("TIMESTAMP'"):
            return datetime.datetime.fromisoformat(obj.strip("TIMESTAMP''"))
        return dateutil.parser.parse(obj)


def do_brackets_sanity_check(s: str) -> bool:
    brackets_pairs = {'(': ')', '[': ']', '{': '}'}
    open_brackets = brackets_pairs.keys()
    close_brackets = brackets_pairs.values()
    stack = []
    for c in s:
        if c in open_brackets:
            stack.append(c)
        elif c in close_brackets:
            if not stack:
                return False
            if brackets_pairs[stack.pop()] != c:
                return False
    return not stack


class ArrayType(BaseType):
    """
    Represents values comprising a sequence of elements with the type of elementType.
    The array type supports sequences of any length greater or equal to 0.
    """

    def __init__(self, element_type: 'BaseType'):
        self.element_type = element_type
        self._is_element_type_complex = element_type.__class__ in ComplexDataTypes

    def __str__(self):
        return f'ARRAY<{self.element_type}>'

    def fit(self, obj) -> bool:
        if isinstance(obj, list) and all([self.element_type.fit(val) for val in obj]):
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'ARRAY({", ".join(map(str, [self.element_type.serialize(val, unsafe=True) for val in obj]))})'

    @staticmethod
    def _split_first_layer(serialized_array: str):
        depth = 0
        current_element_start_idx = None

        for i in range(len(serialized_array)):
            if serialized_array[i] == '(':
                depth += 1
                if current_element_start_idx is None:
                    current_element_start_idx = i + 1
            elif serialized_array[i] == ')':
                depth -= 1
                if depth == 1:
                    current_element_end_idx = i + 1
                    yield serialized_array[current_element_start_idx:current_element_end_idx]
                    current_element_start_idx = None
            elif serialized_array[i] in (',', ' '):
                if depth == 1:
                    current_element_start_idx = i + 1

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, np.ndarray):
            return [self.element_type.deserialize(val) for val in obj.tolist()]
        if isinstance(obj, list):
            return [self.element_type.deserialize(val) for val in obj]
        if isinstance(obj, str) and obj.startswith('ARRAY(') and obj.endswith(')') and do_brackets_sanity_check(obj):
            if obj == 'ARRAY()':
                return []
            if not self._is_element_type_complex:
                return [self.element_type.deserialize(val.strip()) for val in obj.strip('ARRAY()').split(',')]

            return [self.element_type.deserialize(val.strip()) for val in self._split_first_layer(obj)]

        super().raise_uncast_exception(obj)


class BinaryType(BaseType):
    """
    Represents byte sequence values.
    The type supports byte sequences of any length greater or equal to 0.
    """

    def fit(self, obj) -> bool:
        if isinstance(obj, (str, bytes)):
            return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f"X'{obj.encode().hex().upper()}'" if isinstance(obj, str) else f"X'{obj.hex().upper()}'"

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str):
            # At first need to remove support chars. Pattern is X'<hex_string>'
            return bytearray.fromhex(obj.replace("'", '')[1:])
        return obj


class BooleanType(BaseType):
    """
    Represents Boolean values.
    The type supports true and false values.
    """

    def fit(self, obj) -> bool:
        return isinstance(obj, (bool, np.bool_))

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return str(bool(obj)).lower()

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, bool):
            return obj
        if isinstance(obj, str):
            if obj.lower() == 'true':
                return True
            if obj.lower() == 'false':
                return False
        return obj


class IntervalType(BaseType):
    """
    Represents intervals of time either on a scale of seconds or months.
    A year-month interval has a maximal range of +/- 178,956,970 years and 11 months.
    A day-time interval has a maximal range of +/- 106,751,991 days, 23 hours, 59 minutes, and 59.999999 seconds.
    """

    def fit(self, obj) -> bool:
        logger.warning('INTERVAL type is not yet fully supported. Will be used as-is!')
        try:
            str(obj)
            return True
        except ValueError:
            return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return f'INTERVAL {obj}'

    @nullable
    def deserialize(self, obj):
        return obj


class MapType(BaseType):
    """
    Represents values comprising a set of key-value pairs.
    The map type supports maps of any cardinality greater or equal to 0.
    The keys must be unique and not be NULL.
    MAP is not a comparable data type.
    """

    def __init__(self, key_type: 'BaseType', value_type: 'BaseType'):
        self.key_type = key_type
        self.value_type = value_type

    def __str__(self):
        return f'MAP<{self.key_type}, {self.value_type}>'

    def fit(self, obj) -> bool:
        if isinstance(obj, dict):
            pairs = obj.items()
            if all([self.key_type.fit(key) and self.value_type.fit(val) for key, val in pairs]):
                return True
        return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        if isinstance(obj, dict):
            obj = tuple([item for pair in obj.items() for item in pair])
        return f'MAP{obj}'

    @nullable
    def deserialize(self, obj):
        logger.warning('MAP type with complex inner types is not supported for deserialization. Will be used as-is!')
        if isinstance(obj, str) and obj.startswith('MAP'):
            arr = list(map(json.loads, obj.strip('MAP()').split(', ')))
            if len(arr) % 2:
                raise DbxIOTypeError(f'Passed map with odd number of elements, got {obj}')
            return dict(zip(arr[::2], arr[1::2]))
        return dict(obj)


class StringType(BaseType):
    """
    The type supports character sequences of any length greater or equal to 0.
    """

    def fit(self, obj) -> bool:
        return isinstance(obj, str)

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        str_obj = json.dumps(str(obj))
        return f'{str_obj}'

    @nullable
    def deserialize(self, obj):
        return str(obj).replace('"', '')


class StructType(BaseType):
    """
    Represents values with the structure described by a sequence of fields.
    The type supports any number of fields greater or equal to 0.
    """

    def __init__(self, fields: list[tuple[str, 'BaseType']]):
        self.fields = fields

    def __str__(self):
        return f'STRUCT<{", ".join([f"{name} {type_}" for name, type_ in self.fields])}>'

    def fit(self, obj) -> bool:
        if not isinstance(obj, dict):
            return False
        if len(obj) != len(self.fields):
            return False
        for name, type_ in self.fields:
            if name not in obj:
                return False
            if not type_.fit(obj[name]):
                return False
        return True

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        sql_named_struct_values = []
        for name, type_ in self.fields:
            sql_named_struct_values.append(f'"{name}"')
            sql_named_struct_values.append(str(type_.serialize(obj[name], unsafe=True)))

        return f'named_struct({",".join(sql_named_struct_values)})'

    @nullable
    def deserialize(self, obj):
        return obj


class NullType(BaseType):
    def fit(self, obj) -> bool:
        return obj is None

    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return 'NULL'

    def deserialize(self, obj):
        return None


class JSONType(BaseType):
    def __str__(self):
        return str(StringType())

    def fit(self, obj) -> bool:
        try:
            json.dumps(obj)
            return True
        except TypeError:
            return False

    @nullable
    def serialize(self, obj, unsafe: bool = False):
        self.try_fit(obj, unsafe)
        return json.dumps(json.dumps(obj))

    @nullable
    def deserialize(self, obj):
        if isinstance(obj, str):
            return json.loads(obj)
        if isinstance(obj, dict):
            return obj
        raise DbxIOTypeError(f'Cannot cast value `{obj}` to JSON')


class GroupsPrimaryDataTypes:
    INTEGER = (IntType, BigIntType, DecimalType)
    FLOAT = (FloatType, DoubleType)
    DATETIME = (DateType, TimestampType)


ComplexDataTypes = (ArrayType, MapType, StructType)


def as_dbxio_type(value: Any) -> BaseType:
    if isinstance(value, bool) and BooleanType().fit(value):
        return BooleanType()
    if isinstance(value, (int, np.integer)):
        for int_type in GroupsPrimaryDataTypes.INTEGER:
            type_ = int_type()
            if type_.fit(value):
                return type_
    if isinstance(value, (int, float, np.floating)):
        for float_type in GroupsPrimaryDataTypes.FLOAT:
            if float_type().fit(value):
                return float_type()
    if isinstance(value, bytes) and BinaryType().fit(value):
        return BinaryType()
    if isinstance(value, datetime.datetime):
        for dtm_type in GroupsPrimaryDataTypes.DATETIME:
            if dtm_type().fit(value):
                return dtm_type()
    if isinstance(value, list):
        value_type = as_dbxio_type(value[0])
        if ArrayType(value_type).fit(value):
            return ArrayType(value_type)
    if isinstance(value, dict):
        key_type = as_dbxio_type(next(iter(value.keys())))
        value_type = as_dbxio_type(next(iter(value.values())))
        if MapType(key_type=key_type, value_type=value_type).fit(value):
            return MapType(key_type=key_type, value_type=value_type)
    if isinstance(value, dict):
        fields = [(key, as_dbxio_type(val)) for key, val in value.items()]
        if StructType(fields).fit(value):
            return StructType(fields)
    if StringType().fit(value):
        return StringType()
    raise DbxIOTypeError(f"Couldn't determine dbxio type for value {value} of type {type(value)}")


dbxio2pa_types_mapping = {
    str(TinyIntType()): pa.int8(),
    str(SmallIntType()): pa.int16(),
    str(IntType()): pa.int32(),
    str(BigIntType()): pa.int64(),
    str(FloatType()): pa.float32(),
    str(DoubleType()): pa.float64(),
    str(DateType()): pa.date32(),
    str(TimestampType()): pa.timestamp('ns'),
    str(BinaryType()): pa.binary(),
    str(BooleanType()): pa.bool_(),
    str(StringType()): pa.string(),
    str(JSONType()): pa.string(),
    str(NullType()): pa.null(),
}


def convert_dbxio_type_to_pa_type(dbxio_type: BaseType) -> pa.DataType:
    if isinstance(dbxio_type, ArrayType):
        return pa.list_(convert_dbxio_type_to_pa_type(dbxio_type.element_type))
    if isinstance(dbxio_type, MapType):
        return pa.map_(
            convert_dbxio_type_to_pa_type(dbxio_type.key_type), convert_dbxio_type_to_pa_type(dbxio_type.value_type)
        )
    if isinstance(dbxio_type, StructType):
        return pa.struct([pa.field(name, convert_dbxio_type_to_pa_type(type_)) for name, type_ in dbxio_type.fields])
    if isinstance(dbxio_type, DecimalType):
        return pa.decimal128(dbxio_type.precision, dbxio_type.scale)

    try:
        return dbxio2pa_types_mapping[str(dbxio_type)]
    except KeyError:
        raise DbxIOTypeError(f'Cannot convert dbxio type {dbxio_type} to pyarrow type')



from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dbxio.core.cloud.client.object_storage import ObjectStorageClient


def blobs_gc(blobs: list[str], object_storage_client: 'ObjectStorageClient'):
    for blob in blobs:
        object_storage_client.try_delete_blob(blob)


@contextmanager
def blobs_registries(
    object_storage_client: 'ObjectStorageClient',
    keep_blobs: bool = False,
    keep_metablobs: bool = False,
):
    """
    Context manager to manage the blobs and metablobs lists.
    """
    blobs: list[str] = []
    metablobs: list[str] = []
    try:
        yield blobs, metablobs
    finally:
        if not keep_blobs:
            blobs_gc(blobs, object_storage_client)
        if not keep_metablobs:
            blobs_gc(metablobs, object_storage_client)



from dbxio.delta.parsers import infer_schema
from dbxio.delta.table import Materialization, Table, TableAttributes, TableFormat, TableType
from dbxio.delta.table_commands import (
    bulk_write_local_files,
    bulk_write_table,
    create_table,
    drop_table,
    exists_table,
    get_comment_on_table,
    get_tags_on_table,
    merge_table,
    read_table,
    save_table_to_files,
    set_comment_on_table,
    set_tags_on_table,
    unset_comment_on_table,
    unset_tags_on_table,
    write_table,
)
from dbxio.delta.table_schema import TableSchema

__all__ = [
    'Materialization',
    'Table',
    'TableFormat',
    'TableSchema',
    'TableType',
    'TableAttributes',
    'bulk_write_local_files',
    'bulk_write_table',
    'create_table',
    'drop_table',
    'exists_table',
    'infer_schema',
    'merge_table',
    'read_table',
    'write_table',
    'save_table_to_files',
    'get_comment_on_table',
    'get_tags_on_table',
    'set_comment_on_table',
    'set_tags_on_table',
    'unset_comment_on_table',
    'unset_tags_on_table',
]



from dbxio.volume.volume_commands import (
    Volume,
    create_volume,
    download_volume,
    get_comment_on_volume,
    get_tags_on_volume,
    set_comment_on_volume,
    set_tags_on_volume,
    unset_comment_on_volume,
    unset_tags_on_volume,
    write_volume,
)

__all__ = [
    'Volume',
    'create_volume',
    'download_volume',
    'write_volume',
    'get_comment_on_volume',
    'set_comment_on_volume',
    'unset_comment_on_volume',
    'get_tags_on_volume',
    'set_tags_on_volume',
    'unset_tags_on_volume',
]



import json
import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Union

import attrs
import jinja2
from databricks.sdk.service.sql import StatementParameterListItem

from dbxio.sql.types import as_dbxio_type
from dbxio.utils.databricks import ClusterType
from dbxio.utils.logging import get_logger

QUERY_PARAMS_TYPE = Union[dict[str, Any], list[StatementParameterListItem], None]

logger = get_logger()


def log_query(built_query: str) -> None:
    logger.info('Executing query:\n%s', built_query)


def cast_params_to_statement_api(params: Union[dict, None]) -> Union[list[StatementParameterListItem], None]:
    if params is None:
        return None

    new_params = []
    for key, value in params.items():
        dbxio_type = as_dbxio_type(value)
        new_params.append(StatementParameterListItem(name=key, value=value, type=str(dbxio_type)))

    return new_params


class BaseDatabricksQuery(metaclass=ABCMeta):
    @abstractmethod
    def build(self, cluster_type: ClusterType) -> tuple[str, QUERY_PARAMS_TYPE]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def params(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def query(self) -> str:
        raise NotImplementedError()


@attrs.define
class ConstDatabricksQuery(BaseDatabricksQuery):
    query: str
    params = None

    def build(self, cluster_type: ClusterType) -> tuple[str, None]:
        return self.query, None


@attrs.define
class JinjaDatabricksQuery(BaseDatabricksQuery):
    query: str
    params: dict[str, Any]

    def __attrs_post_init__(self):
        warnings.warn(
            'JinjaDatabricksQuery is deprecated and will be removed in the future. '
            'Use ParametrizedDatabricksQuery instead',
            DeprecationWarning,
        )

    def build(self, cluster_type: ClusterType) -> tuple[str, dict[str, Any]]:
        jinja_env = jinja2.Environment()
        query_template = jinja_env.from_string(self.query)
        for k, v in self.params.items():
            if isinstance(v, (list, tuple, set)):
                self.params[k] = f"({','.join(map(json.dumps, v))})"
        return query_template.render(**self.params), self.params


@attrs.define
class ParametrizedDatabricksQuery(BaseDatabricksQuery):
    """
    A class for parametrized Databricks SQL queries.
    Only primitive types are supported as parameters.
    Lists, tuples, and sets are converted to comma-separated strings.
    """

    query: str
    params: dict[str, Any]

    def build(self, cluster_type: ClusterType) -> tuple[str, QUERY_PARAMS_TYPE]:
        built_query = self.query

        if cluster_type == ClusterType.SQL_WAREHOUSE:
            return built_query, cast_params_to_statement_api(self.params)

        return built_query, self.params



import re
from abc import ABC, abstractmethod
from io import IOBase
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO, Iterator, Optional, Type, Union

if TYPE_CHECKING:
    from dbxio.core.cloud import CloudProvider


class ObjectStorageClient(ABC):
    url_regex: re.Pattern

    # azure properties
    container_name: str
    storage_name: str
    domain_name: Optional[str]
    blobs_path: Optional[str]

    # s3/gs properties
    bucket_name: str
    object_key: str

    def __init__(self, **kwargs):
        raise NotImplementedError

    @property
    @abstractmethod
    def scheme(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def cloud_provider_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def credential_provider(self):
        raise NotImplementedError

    @abstractmethod
    def to_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def try_delete_blob(self, blob_name: str):
        raise NotImplementedError

    @abstractmethod
    def list_blobs(self, prefix: Optional[str] = None, **kwargs) -> Iterator:
        raise NotImplementedError

    @abstractmethod
    def download_blob(self, blob_name: str) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def download_blob_to_file(self, blob_name: str, file_path: Union[str, Path]) -> None:
        raise NotImplementedError

    @abstractmethod
    def upload_blob(self, blob_name: str, data: Union[bytes, IOBase, BinaryIO], overwrite: bool = False, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def lock_blob(self, blob_name: str, force: bool = False):
        raise NotImplementedError

    @abstractmethod
    def break_lease(self, blob_name: str) -> None:
        raise NotImplementedError

    @staticmethod
    def _get_storage_impl(
        value: 'Union[str | CloudProvider]',
        by_scheme: bool = False,
        by_cloud_provider: bool = False,
    ) -> 'Type[ObjectStorageClient]':
        from dbxio.core.cloud.aws.object_storage import _S3StorageClientImpl
        from dbxio.core.cloud.azure.object_storage import _AzureBlobStorageClientImpl
        from dbxio.core.cloud.gcp.object_storage import _GCStorageClientImpl

        assert by_scheme or by_cloud_provider, 'Either by_scheme or by_cloud_provider should be set to True'

        attr_name = 'scheme' if by_scheme else 'cloud_provider_name'
        if value == getattr(_AzureBlobStorageClientImpl, attr_name):
            return _AzureBlobStorageClientImpl
        if value == getattr(_S3StorageClientImpl.scheme, attr_name):
            return _S3StorageClientImpl
        if value == getattr(_GCStorageClientImpl.scheme, attr_name):
            return _GCStorageClientImpl

        raise ValueError(f'Unsupported storage type: {value}')

    @classmethod
    def from_url(cls, storage_url: str) -> 'ObjectStorageClient':
        scheme, url = storage_url.split('://')
        _storage_impl = cls._get_storage_impl(scheme, by_scheme=True)
        _match = _storage_impl.url_regex.match(url)
        if _match:
            return _storage_impl(**_match.groupdict())

        raise ValueError(f'Invalid Storage URL: {url}')

    @classmethod
    def from_storage_options(cls, cloud_provider: 'CloudProvider', **storage_options) -> 'ObjectStorageClient':
        _storage_impl = cls._get_storage_impl(cloud_provider, by_cloud_provider=True)
        return _storage_impl(**storage_options)



import logging
import os.path
import uuid
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Union

from databricks.sql import ServerOperationError

from dbxio.blobs.block_upload import upload_file
from dbxio.blobs.parquet import create_pa_table, create_tmp_parquet, pa_table2parquet
from dbxio.core.cloud.client.object_storage import ObjectStorageClient
from dbxio.delta.parsers import infer_schema
from dbxio.delta.table import Table, TableFormat
from dbxio.sql.query import ConstDatabricksQuery
from dbxio.sql.results import _FutureBaseResult
from dbxio.utils.blobs import blobs_registries
from dbxio.utils.logging import get_logger

if TYPE_CHECKING:
    from dbxio.core import DbxIOClient

logger = get_logger()


def exists_table(table: Union[str, Table], client: 'DbxIOClient') -> bool:
    """
    Checks if table exists in the catalog. Tries to read one record from the table.
    """
    table = Table.from_obj(table)

    try:
        next(read_table(table, limit_records=1, client=client))
        return True
    except ServerOperationError:
        return False


def create_table(table: Union[str, Table], client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Creates a table in the catalog.
    If a table already exists, it does nothing.
    Query pattern:
        CREATE TABLE IF NOT EXISTS <table_identifier> (col1 type1, col2 type2, ...)
        [USING <table_format> LOCATION <location>]
        [PARTITIONED BY (col1, col2, ...)]
    """
    dbxio_table = Table.from_obj(table)

    schema_sql = ','.join([f'`{col_name}` {col_type}' for col_name, col_type in dbxio_table.schema.as_dict().items()])
    query = f'CREATE TABLE IF NOT EXISTS {dbxio_table.safe_table_identifier} ({schema_sql})'
    if loc := dbxio_table.attributes.location:
        query += f" USING {dbxio_table.table_format.name} LOCATION '{loc}'"
    if part := dbxio_table.attributes.partitioned_by:
        query += f" PARTITIONED BY ({','.join(part)})"

    return client.sql(query)


def drop_table(table: Union[str, Table], client: 'DbxIOClient', force: bool = False) -> _FutureBaseResult:
    """
    Drops a table from the catalog.
    Operation can be forced by setting force=True.
    """
    dbxio_table = Table.from_obj(table)

    force_mark = 'IF EXISTS ' if force else ''
    drop_sql = f'DROP TABLE {force_mark}{dbxio_table.safe_table_identifier}'

    return client.sql(drop_sql)


def read_table(
    table: Union[str, Table],
    client: 'DbxIOClient',
    columns_subset: list[str] = None,
    *,
    distinct: bool = False,
    limit_records: int = None,
) -> Iterator[Dict[str, Any]]:
    """
    Reads data from the table.
    Data amount can be limited by columns_subset, distinct or limit_records options.
    Returns a generator of records and applies schema if it's present.
    """
    logging.getLogger('thrift_backend.py').setLevel('INFO')  # HACK: disable logging from databricks.sql
    dbxio_table = Table.from_obj(table)

    # FIXME use sql-builder
    sql_columns_subset = ','.join([f'`{col}`' for col in columns_subset]) if columns_subset else '*'
    sql_distinct = 'DISTINCT ' if distinct else ''
    _sql_query = f'SELECT {sql_distinct}{sql_columns_subset} FROM {dbxio_table.safe_table_identifier}'
    if limit_records:
        _sql_query += f' LIMIT {limit_records}'

    with client.sql(_sql_query) as fq:
        for records in fq:
            records = records if isinstance(records, list) else [records]
            for record in records:
                if dbxio_table.schema:
                    record = dbxio_table.schema.apply(record)
                yield record


def save_table_to_files(
    table: Union[str, Table],
    client: 'DbxIOClient',
    results_path: str,
    max_concurrency: int = 1,
) -> Path:
    """
    Saves table content to specified path.
    Returns path to the results.

    Data can be saved in multiple files.
    Sql driver determines the number of files.
    """
    dbxio_table = Table.from_obj(table)

    sql_read_query = f'select * from {dbxio_table.safe_table_identifier}'
    return client.sql_to_files(sql_read_query, results_path=results_path, max_concurrency=max_concurrency)


def write_table(
    table: Union[str, Table],
    new_records: Union[Iterator[Dict], List[Dict]],
    client: 'DbxIOClient',
    append: bool = True,
) -> _FutureBaseResult:
    """
    Writes new records to the table using a direct SQL statement.
    Function is relatively fast on small datasets but can be slow or even fail on large datasets.

    For all use cases, consider using bulk_write_table or bulk_write_local_files functions.
    Databricks does not support sql queries larger than 50 Mb.
    """
    dbxio_table = Table.from_obj(table)

    new_records = iter(new_records)
    first_record = next(new_records)

    if dbxio_table.schema is None:
        dbxio_table.schema = infer_schema(first_record)
    schema = dbxio_table.schema.as_dict()

    create_table(dbxio_table, client=client).wait()

    column_names = dbxio_table.schema.columns
    input_way = 'INTO' if append else 'OVERWRITE'
    new_values_sql_format = [
        f'( {",".join([schema[col_name].serialize(first_record.get(col_name)) for col_name in column_names])} )'
    ]
    new_values_sql_format += [
        f'( {",".join([schema[col_name].serialize(record.get(col_name)) for col_name in column_names])} )'
        for record in new_records
    ]

    _sql_query = (
        f"INSERT {input_way} {dbxio_table.safe_table_identifier} ({','.join(column_names)}) "
        f"VALUES {','.join(map(str, new_values_sql_format))}"
    )

    query_size_in_bytes = len(_sql_query.encode('utf-8'))
    if query_size_in_bytes > 10 * 2**20:
        # TODO: raise exception here
        logger.warning(
            f'Query size is {query_size_in_bytes / 2 ** 20} MB. '
            f'Please consider using bulk_write_table function instead of write_table.'
            f'In further versions of dbxio this function will be deprecated and removed.'
        )

    return client.sql(_sql_query)


def copy_into_table(
    client: 'DbxIOClient',
    table: Table,
    blob_path: str,
    table_format: TableFormat,
    abs_name: str,
    abs_container_name: str,
    include_files_pattern: bool = False,
) -> None:
    """
    Copy data from blob storage into the table. All files that match the pattern *.{table_format} will be copied.
    """
    _copy_into_query = dedent(f"""
    COPY INTO {table.safe_table_identifier}
    FROM "abfss://{abs_container_name}@{abs_name}.dfs.core.windows.net/{blob_path}"
    FILEFORMAT = {table_format.value}
    """)
    _pattern = f"PATTERN = '*.{table_format.value.lower()}'" if include_files_pattern else ''
    _options = dedent("""
    FORMAT_OPTIONS ("mergeSchema" = "true")
    COPY_OPTIONS ("mergeSchema" = "true")
    """)

    sql_copy_into_query = ConstDatabricksQuery(f'{_copy_into_query} {_pattern} {_options}')
    client.sql(sql_copy_into_query).wait()


def bulk_write_table(
    table: Union[str, Table],
    new_records: Union[Iterator[Dict], List[Dict]],
    client: 'DbxIOClient',
    abs_name: str,
    abs_container_name: str,
    append: bool = True,
) -> None:
    """
    Bulk write table using parquet format and CLONE INTO statement.
    Function requires blob storage to store temporary parquet file.
    """
    dbxio_table = Table.from_obj(table)
    stream = iter(new_records)
    first_record = next(stream)
    dbxio_table.schema = dbxio_table.schema or infer_schema(first_record)

    columnar_table: dict[str, list] = {col_name: [] for col_name in dbxio_table.schema.columns}
    for col_name in dbxio_table.schema.columns:
        columnar_table[col_name].append(first_record.get(col_name))
    for record in stream:
        for col_name in dbxio_table.schema.columns:
            columnar_table[col_name].append(record.get(col_name))

    pa_table = create_pa_table(columnar_table, schema=dbxio_table.schema)
    pa_table_as_bytes = pa_table2parquet(pa_table)

    object_storage = ObjectStorageClient.from_storage_options(
        cloud_provider=client.settings.cloud_provider,
        storage_name=abs_name,
        container_name=abs_container_name,
        credential_provider=client.credential_provider.az_cred_provider,
    )

    with create_tmp_parquet(pa_table_as_bytes, dbxio_table.table_identifier, object_storage) as tmp_path:
        if not append:
            drop_table(dbxio_table, client=client, force=True).wait()
        create_table(dbxio_table, client=client).wait()

        copy_into_table(
            client=client,
            table=dbxio_table,
            table_format=TableFormat.PARQUET,
            blob_path=tmp_path,
            abs_name=abs_name,
            abs_container_name=abs_container_name,
        )


def bulk_write_local_files(
    table: Table,
    path: str,
    table_format: TableFormat,
    client: 'DbxIOClient',
    abs_name: str,
    abs_container_name: str,
    append: bool = True,
    force: bool = False,
    max_concurrency: int = 1,
) -> None:
    """
    Write data from local files to the table.
    """
    assert table.schema, 'Table schema is required for bulk_write_local_files function'

    p = Path(path)
    files = p.glob(f'*.{table_format.value.lower()}') if p.is_dir() else [path]

    operation_uuid = str(uuid.uuid4())
    object_storage = ObjectStorageClient.from_storage_options(
        cloud_provider=client.settings.cloud_provider,
        storage_name=abs_name,
        container_name=abs_container_name,
        credential_provider=client.credential_provider.az_cred_provider,
    )
    with blobs_registries(object_storage_client=object_storage) as (blobs, metablobs):
        for filename in files:
            upload_file(
                filename,  # type: ignore
                p,
                object_storage_client=object_storage,
                operation_uuid=operation_uuid,
                blobs=blobs,
                metablobs=metablobs,
                max_concurrency=max_concurrency,
                force=force,
            )

        if not append:
            drop_table(table, client=client, force=True).wait()
        create_table(table, client=client).wait()

        common_blob_path = str(os.path.commonpath(blobs))
        include_files_pattern = len(blobs) > 1
        copy_into_table(
            client=client,
            table=table,
            table_format=table_format,
            blob_path=common_blob_path,
            include_files_pattern=include_files_pattern,
            abs_name=abs_name,
            abs_container_name=abs_container_name,
        )


def merge_table(
    table: 'Union[str , Table]',
    new_records: 'Union[Iterator[Dict] , List[Dict]]',
    partition_by: 'Union[str , List[str]]',
    client: 'DbxIOClient',
) -> None:
    """
    Merge new data into table. Use this function only if you have partitioning. Without partitioning, it's the same as
    append write operation.
    Function always waits till the end of deleting tmp table
    """
    dbxio_table = Table.from_obj(table)
    tmp_table = Table(table_identifier=f'{dbxio_table.table_identifier}__dbxio_tmp', schema=dbxio_table.schema)

    write_table(tmp_table, new_records, client=client, append=False).wait()

    _destination_alias = 'DBXIO_DESTINATION'
    _source_alias = 'DBXIO_SOURCE'
    if isinstance(partition_by, str):
        partition_by = [partition_by]
    partition_by_sql_statement = ' AND '.join(
        [f'{_source_alias}.{part_col} == {_destination_alias}.{part_col}' for part_col in partition_by]
    )

    merge_sql_query = dedent(f"""
        MERGE INTO {dbxio_table.safe_table_identifier} AS {_destination_alias}
        USING {tmp_table.safe_table_identifier} AS {_source_alias}
        ON {partition_by_sql_statement}

        WHEN MATCHED THEN UPDATE SET *
        WHEN NOT MATCHED THEN INSERT *
        """)
    try:
        client.sql(merge_sql_query).wait()
    finally:
        drop_table(tmp_table, client=client, force=True).wait()


def set_comment_on_table(
    table: 'Union[str , Table]',
    comment: Union[str, None],
    client: 'DbxIOClient',
) -> _FutureBaseResult:
    """
    Sets a comment on a table.
    Description comment supports Markdown.

    If the comment is None, the comment will be removed.
    """
    dbxio_table = Table.from_obj(table)
    set_comment_query = dedent(
        f"""COMMENT ON TABLE {dbxio_table.safe_table_identifier} IS {f'"{comment}"' if comment else 'NULL'}"""
    )

    return client.sql(set_comment_query)


def unset_comment_on_table(table: 'Union[str , Table]', client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Unsets the comment on a table.
    """
    return set_comment_on_table(table=table, comment=None, client=client)


def get_comment_on_table(table: 'Union[str , Table]', client: 'DbxIOClient') -> Union[str, None]:
    """
    Returns the comment on a table.
    """
    dbxio_table = Table.from_obj(table)
    table_catalog, table_schema, table_name = dbxio_table.table_identifier.split('.')
    information_schema_query = dedent(f"""
    select comment
    from system.information_schema.tables
    where
        table_catalog = '{table_catalog}'
        and table_schema = '{table_schema}'
        and table_name = '{table_name}'
    """)

    with client.sql(information_schema_query) as result:
        for row in result:
            return row['comment']

    return None


def set_tags_on_table(table: 'Union[str , Table]', tags: dict[str, str], client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Sets tags on a table.
    Each tag is a key-value pair of strings.
    """
    assert tags, 'Tags must be a non-empty dictionary.'

    dbxio_table = Table.from_obj(table)
    set_tags_query = dedent(f"""
    ALTER TABLE {dbxio_table.safe_table_identifier}
    SET TAGS ({', '.join([f'"{k}" = "{v}"' for k, v in tags.items()])})
    """)

    return client.sql(set_tags_query)


def unset_tags_on_table(table: 'Union[str , Table]', tags: list[str], client: 'DbxIOClient') -> _FutureBaseResult:
    """
    Unsets tags on a table.
    """
    assert tags, 'Tags must be a non-empty list.'

    dbxio_table = Table.from_obj(table)
    unset_tags_query = dedent(f"""
    ALTER TABLE {dbxio_table.safe_table_identifier}
    UNSET TAGS ({', '.join([f'"{tag}"' for tag in tags])})
    """)

    return client.sql(unset_tags_query)


def get_tags_on_table(table: 'Union[str , Table]', client: 'DbxIOClient') -> dict[str, str]:
    """
    Returns the tags on a table.
    """
    dbxio_table = Table.from_obj(table)
    catalog_name, schema_name, table_name = dbxio_table.table_identifier.split('.')
    information_schema_query = dedent(f"""
    select tag_name, tag_value
    from system.information_schema.table_tags
    where
        catalog_name = '{catalog_name}'
        and schema_name = '{schema_name}'
        and table_name = '{table_name}'
    """)

    tags = {}
    with client.sql(information_schema_query) as result:
        for row in result:
            tags[row['tag_name']] = row['tag_value']

    return tags



import hashlib
from pathlib import Path
from typing import TYPE_CHECKING, Union

from azure.storage.blob import BlobType

from dbxio.core.cloud.exceptions import BlobModificationError
from dbxio.utils.logging import get_logger

if TYPE_CHECKING:
    from dbxio.core.cloud.client.object_storage import ObjectStorageClient

_HASHSUM_SUFFIX = '_HASHSUM'
_SUCCESS_SUFFIX = '_SUCCESS'
_LOCK_SUFFIX = '_LOCK'

logger = get_logger()


def _local_path_to_blob_name(file_path: Path, local_path: Path, operation_uuid: str) -> str:
    relative_path = file_path.relative_to(local_path) if file_path != local_path else file_path.name
    return f'{operation_uuid}/{relative_path}'


def _get_file_hash(file_path: Path) -> str:
    sha256 = hashlib.sha256()
    with open(file_path, 'rb') as f:
        for byte_block in iter(lambda: f.read(4096), b''):
            sha256.update(byte_block)
    return sha256.hexdigest()


def _blob_exists(object_storage_client: 'ObjectStorageClient', blob_name: str, target_hashsum: str) -> bool:
    """
    Checks that blob exists in the container. It's also required to check that hashsums are equal and file with
    suffix _SUCCESS exists.
    """
    blobs = [blob.name for blob in object_storage_client.list_blobs(prefix=blob_name)]
    if f'{blob_name}{_SUCCESS_SUFFIX}' not in blobs or f'{blob_name}{_HASHSUM_SUFFIX}' not in blobs:
        logger.debug(f'Blob {blob_name} does not exist (no SUCCESS or HASHSUM file)')
        return False
    saved_hashsum = object_storage_client.download_blob(f'{blob_name}{_HASHSUM_SUFFIX}').decode()
    return saved_hashsum == target_hashsum


def _lock_blob(blob_name: str, object_storage_client: 'ObjectStorageClient', force: bool = False):
    """
    Locks blob by creating a lease on <blob_name>_LOCK file.
    If any other process tries to upload the same file, it will fail to acquire the lease.
    If force is True, it will break the lease and acquire it again.
    """
    try:
        object_storage_client.upload_blob(f'{blob_name}{_LOCK_SUFFIX}', b'', overwrite=True)
    except BlobModificationError:
        object_storage_client.lock_blob(f'{blob_name}{_LOCK_SUFFIX}', force=force)

    logger.debug(f'Lock is acquired for {blob_name}')


def upload_file(
    path: Union[str, Path],
    local_path: Union[str, Path],
    object_storage_client: 'ObjectStorageClient',
    blobs: list[str],
    metablobs: list[str],
    operation_uuid: str,
    max_concurrency: int = 1,
    force: bool = False,
) -> str:
    """
    Uploads file to the Azure Blob Storage container with guarantees that only one process can upload the same file at
    the same time.
    It also checks that the file is not uploaded yet by comparing hashsums.
    """
    path = Path(path)
    local_path = Path(local_path)
    logger.debug(f'Using {max_concurrency} threads for uploading {path}')

    file_hash = _get_file_hash(path)
    blob_name = _local_path_to_blob_name(path, local_path, operation_uuid)
    metablobs.append(f'{blob_name}{_LOCK_SUFFIX}')
    metablobs.append(f'{blob_name}{_SUCCESS_SUFFIX}')
    metablobs.append(f'{blob_name}{_HASHSUM_SUFFIX}')

    blobs.append(blob_name)

    if _blob_exists(object_storage_client, blob_name, file_hash):
        return blob_name

    _lock_blob(blob_name, object_storage_client, force=force)

    with open(path, 'rb') as data:
        object_storage_client.upload_blob(
            blob_name,
            data,
            blob_type=BlobType.BLOCKBLOB,
            overwrite=True,
            max_concurrency=int(max_concurrency),
        )

    object_storage_client.upload_blob(f'{blob_name}{_SUCCESS_SUFFIX}', b'', overwrite=True)
    logger.debug(f'SUCCESS file is uploaded for {blob_name}')
    object_storage_client.upload_blob(f'{blob_name}{_HASHSUM_SUFFIX}', file_hash.encode(), overwrite=True)
    logger.debug(f'HASHSUM file is uploaded for {blob_name}')

    logger.info(f'Successfully uploaded {path} to {object_storage_client.to_url()}')

    return blob_name



from dbxio.core import *  # noqa: F403
from dbxio.delta import *  # noqa: F403
from dbxio.sql import *  # noqa: F403
from dbxio.utils import *  # noqa: F403
from dbxio.volume import *  # noqa: F403

__version__ = '0.2.1'  # single source of truth



import re
from typing import Optional

import attrs

from dbxio.core.cloud.client.object_storage import ObjectStorageClient
from dbxio.core.cloud.gcp import GCP_ClOUD_PROVIDER_NAME


@attrs.define
class _GCStorageClientImpl(ObjectStorageClient):
    bucket_name: str = attrs.field(validator=attrs.validators.instance_of(str))
    object_key: str = attrs.field(validator=attrs.validators.instance_of(str))
    domain_name: Optional[str] = None

    scheme = 'gs'
    cloud_provider_name = GCP_ClOUD_PROVIDER_NAME
    # bucket_name/object_key
    url_regex = re.compile(r'^(?P<bucket_name>[^/]+)/(?P<object_key>.*)$')

    def to_url(self) -> str:
        return f'{self.scheme}://{self.bucket_name}/{self.object_key}'



from dbxio.sql import types
from dbxio.sql.query import (
    ConstDatabricksQuery,
    JinjaDatabricksQuery,
    ParametrizedDatabricksQuery,
    cast_params_to_statement_api,
)
from dbxio.sql.sql_driver import ODBCDriver, SQLDriver, StatementExecutionAPI, get_sql_driver
from dbxio.sql.types import as_dbxio_type

__all__ = [
    'ConstDatabricksQuery',
    'JinjaDatabricksQuery',
    'ParametrizedDatabricksQuery',
    'cast_params_to_statement_api',
    'SQLDriver',
    'ODBCDriver',
    'StatementExecutionAPI',
    'get_sql_driver',
    'types',
    'as_dbxio_type',
]



import os

import attrs

from dbxio.core.cloud import CloudProvider

_DEFAULT_CLOUD_PROVIDER = CloudProvider.AZURE  # TODO: get provider from env or remove this default at all
_SUPPORTED_CLOUD_PROVIDERS = [
    CloudProvider.AZURE,
]

CLOUD_PROVIDER_ENV_VAR = 'CLOUD_PROVIDER'


def _cloud_provider_factory() -> CloudProvider:
    return CloudProvider(os.getenv(CLOUD_PROVIDER_ENV_VAR, _DEFAULT_CLOUD_PROVIDER).lower())


@attrs.define
class Settings:
    cloud_provider: CloudProvider = attrs.field(factory=_cloud_provider_factory)

    @cloud_provider.validator
    def _validate_cloud_provider(self, attribute, value):
        if not isinstance(value, CloudProvider):
            value = CloudProvider(value)
        if value not in _SUPPORTED_CLOUD_PROVIDERS:
            raise ValueError(
                f'Unsupported cloud provider: {value}. Supported providers: {_SUPPORTED_CLOUD_PROVIDERS}. '
                f'Please raise an issue if you want to add a new provider '
                f'(https://github.com/Toloka/dbxio/issues/new/choose)'
            )



from pathlib import Path
from typing import TYPE_CHECKING, Optional

from tenacity import retry, stop_after_attempt, wait_fixed

if TYPE_CHECKING:
    from dbxio.core.cloud.client.object_storage import ObjectStorageClient


@retry(stop=stop_after_attempt(3), wait=wait_fixed(10))
def download_blob_tree(
    object_storage_client: 'ObjectStorageClient', local_path: Path, prefix_path: Optional[str] = None
):
    for blob in object_storage_client.list_blobs(prefix=prefix_path):
        if blob.content_settings.content_type is None:
            # it's a directory, create it
            Path(local_path / blob.name).mkdir(parents=True, exist_ok=True)
            continue

        object_storage_client.download_blob_to_file(blob.name, local_path / blob.name)



import copy
from functools import cache
from typing import Any

try:
    import pydantic.v1 as pydantic
except ModuleNotFoundError:
    import pydantic  # type: ignore

from dbxio.sql.types import BaseType


class ColumnSpec(pydantic.BaseModel):
    name: str
    type: BaseType

    class Config:
        arbitrary_types_allowed = True


class TableSchema:
    """
    Represents a schema of a table.
    Schema is a collection of columns and associated types.
    To create a schema, use factory method `from_obj` with one of the following types:
        - list[dict] - list of dictionaries with keys 'name' and 'type', like
            >>> TableSchema.from_obj([{'name': 'id', 'type': IntType()}, {'name': 'name', 'type': StringType()}])
        - dict - dictionary with column names as keys and types as values, like
            >>> TableSchema.from_obj({'id': IntType(), 'name': StringType()})
    """

    def __init__(self, raw_schema: list[dict[str, BaseType]]):
        self._columns: list[ColumnSpec] = self._check_schema_obj(raw_schema)

    def __getattr__(self, item):
        try:
            column = [col for col in self._columns if col.name == item][0]
        except IndexError:
            raise AttributeError(f'Column {item} not found in schema. Possible columns: {self.columns}')
        return column

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)

        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))

        return result

    @classmethod
    def from_obj(cls, obj):
        if obj is None:
            return None
        if isinstance(obj, list):
            return TableSchema(obj)
        if isinstance(obj, dict):
            return TableSchema([{'name': k, 'type': v} for k, v in obj.items()])
        if isinstance(obj, TableSchema):
            return obj

        raise TypeError(f'Got unexpected obj with type {type(obj)}')

    @staticmethod
    def _check_schema_obj(schema: list[dict]) -> list[ColumnSpec]:
        return pydantic.parse_obj_as(list[ColumnSpec], schema)

    @property
    def columns(self):
        return [col.name for col in self._columns]

    @cache
    def as_dict(self) -> dict[str, BaseType]:
        return {col_spec.name: col_spec.type for col_spec in self._columns}

    def apply(self, record: dict[str, Any]) -> dict[str, Any]:
        return {key: self.as_dict()[key].deserialize(val) for key, val in record.items()}



import uuid
from contextlib import contextmanager
from typing import TYPE_CHECKING, Dict, Iterator, Union

import pyarrow as pa
import pyarrow.parquet as pq

from dbxio.sql.types import convert_dbxio_type_to_pa_type

if TYPE_CHECKING:
    from dbxio.core.cloud.client.object_storage import ObjectStorageClient
    from dbxio.delta.table import Table
    from dbxio.delta.table_schema import TableSchema

ROW_GROUP_SIZE_BYTES = 128 * 2**20


def create_pa_table(parsed_micro_batch: Dict, schema: 'TableSchema') -> pa.Table:
    batch_as_arrays = []
    schema_as_dict = schema.as_dict()
    for col_name, col_values in parsed_micro_batch.items():
        col_type = convert_dbxio_type_to_pa_type(schema_as_dict[col_name])
        batch_as_arrays.append(pa.array(col_values, type=col_type))

    return pa.Table.from_arrays(arrays=batch_as_arrays, names=list(parsed_micro_batch.keys()))


def pa_table2parquet(table: pa.Table) -> bytes:
    stream = pa.BufferOutputStream()
    row_group_size = int(table.num_rows / int(table.nbytes / ROW_GROUP_SIZE_BYTES + 1) + 1)
    pq.write_table(table, stream, flavor={'spark'}, row_group_size=row_group_size)
    return stream.getvalue().to_pybytes()


def arrow_stream2parquet(stream: bytes) -> bytes:
    table = pa.ipc.open_stream(stream).read_all()
    return pa_table2parquet(table)


@contextmanager
def create_tmp_parquet(
    data: bytes,
    table_identifier: Union[str, 'Table'],
    object_storage_client: 'ObjectStorageClient',
) -> Iterator[str]:
    random_part = uuid.uuid4()
    ti = table_identifier if isinstance(table_identifier, str) else table_identifier.table_identifier
    translated_table_identifier = ti.translate(
        str.maketrans('.!"#$%&\'()*+,/:;<=>?@[\\]^`{|}~', '______________________________')
    )
    tmp_path = f'{translated_table_identifier}__dbxio_tmp__{random_part}.parquet'
    object_storage_client.upload_blob(tmp_path, data, overwrite=True)
    try:
        yield tmp_path
    finally:
        object_storage_client.try_delete_blob(tmp_path)



import re
from io import IOBase
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO, Iterator, Optional, Union

import attrs
from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobLeaseClient, BlobServiceClient
from cachetools import TTLCache

from dbxio.core.cloud.azure import AZURE_ClOUD_PROVIDER_NAME
from dbxio.core.cloud.client.object_storage import ObjectStorageClient
from dbxio.core.cloud.exceptions import BlobModificationError
from dbxio.utils.logging import get_logger

if TYPE_CHECKING:
    from dbxio.core.auth import AZ_CRED_PROVIDER_TYPE

logger = get_logger()


@attrs.define
class _AzureBlobStorageClientImpl(ObjectStorageClient):
    container_name: str = attrs.field(validator=attrs.validators.instance_of(str))
    storage_name: str = attrs.field(validator=attrs.validators.instance_of(str))
    blobs_path: str = attrs.field(default='', validator=attrs.validators.instance_of(str))
    domain_name: str = attrs.field(default='dfs.core.windows.net', validator=attrs.validators.instance_of(str))
    credential_provider: 'AZ_CRED_PROVIDER_TYPE' = attrs.field(default=DefaultAzureCredential())
    blob_service_client: BlobServiceClient = attrs.field()

    scheme = 'abfss'
    cloud_provider_name = AZURE_ClOUD_PROVIDER_NAME
    # container_name@storage_name.domain_name/blobs_path
    url_regex = re.compile(
        r'(?P<container_name>[^@]+)@(?P<storage_name>[^.]+)\.(?P<domain_name>[^/]+)/(?P<blobs_path>.*)$'
    )
    _cache: TTLCache = attrs.Factory(lambda: TTLCache(maxsize=1024, ttl=60 * 15))

    @blob_service_client.default
    def _blob_service_client_factory(self) -> BlobServiceClient:
        return BlobServiceClient(
            account_url=f'https://{self.storage_name}.blob.core.windows.net',
            credential=self.credential_provider,
        )

    def to_url(self) -> str:
        return f'{self.scheme}://{self.container_name}@{self.storage_name}.{self.domain_name}/{self.blobs_path}'

    @property
    def account_url(self) -> str:
        return f'https://{self.storage_name}.blob.core.windows.net'

    def list_blobs(self, prefix: Optional[str] = None, **kwargs) -> Iterator:
        container_client = self.blob_service_client.get_container_client(self.container_name)
        return container_client.list_blobs(name_starts_with=prefix, **kwargs)

    def download_blob(self, blob_name: str) -> bytes:
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=blob_name)
        return blob_client.download_blob().readall()

    def download_blob_to_file(self, blob_name: str, file_path: Union[str, Path]) -> None:
        with open(file_path, 'wb') as f:
            f.write(self.download_blob(blob_name))

    def break_lease(self, blob_name: str) -> None:
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=blob_name)
        BlobLeaseClient(client=blob_client).break_lease()  # type: ignore

    def lock_blob(self, blob_name: str, force: bool = False):
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=blob_name)
        if force:
            self.break_lease(blob_name)
        try:
            blob_client.acquire_lease()
        except ResourceExistsError as e:
            raise BlobModificationError(f'Failed to lock blob {blob_name}') from e

    def upload_blob(self, blob_name: str, data: Union[bytes, IOBase, BinaryIO], overwrite: bool = False, **kwargs):
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=blob_name)
        try:
            blob_client.upload_blob(data, **kwargs, overwrite=overwrite)
        except ResourceExistsError as e:
            raise BlobModificationError(f'Failed to upload blob {blob_name}') from e

    def try_delete_blob(self, blob_name: str) -> None:
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=blob_name)
        lease_client = BlobLeaseClient(client=blob_client)  # type: ignore

        try:
            lease_client.break_lease()
        except (ResourceExistsError, ResourceNotFoundError):
            pass

        try:
            blob_client.delete_blob()
        except ResourceNotFoundError:
            pass
